# This file is Copyright 2020 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#

import json
import logging
import lzma
import os
import re
import struct
from pathlib import PureWindowsPath
from typing import Any, Dict, Generator, List, Optional, Tuple, Union
from urllib import parse, request

from volatility3 import symbols
from volatility3.framework import constants, contexts, exceptions, interfaces
from volatility3.framework.automagic import symbol_cache
from volatility3.framework.configuration import requirements
from volatility3.framework.symbols import intermed
from volatility3.framework.symbols.windows import pdbconv

vollog = logging.getLogger(__name__)


class PDBUtility(interfaces.configuration.VersionableInterface):
    """Class to handle and manage all getting symbols based on MZ header"""

    _version = (1, 0, 1)
    _required_framework_version = (2, 0, 0)

    @classmethod
    def symbol_table_from_offset(
        cls,
        context: interfaces.context.ContextInterface,
        layer_name: str,
        offset: int,
        symbol_table_class: str = "volatility3.framework.symbols.intermed.IntermediateSymbolTable",
        config_path: Optional[str] = None,
        progress_callback: constants.ProgressCallback = None,
    ) -> Optional[str]:
        """Produces the name of a symbol table loaded from the offset for an MZ header

        Args:
            context: The context on which to operate
            layer_name: The name of the (contiguous) layer within the context that contains the MZ file
            offset: The offset in the layer at which the MZ file begins
            symbol_table_class: The class to use when constructing the SymbolTable
            config_path: New path for the produced symbol table configuration with the config tree
            progress_callback: Callable called to update ongoing progress

        Returns:
            None if no pdb information can be determined, else returned the name of the loaded symbols for the MZ
        """
        result = cls.get_guid_from_mz(context, layer_name, offset)
        if result is None:
            vollog.debug(f"Could not get GUID for {hex(offset)}")
            return None
        guid, age, pdb_name = result
        if config_path is None:
            config_path = interfaces.configuration.path_join(
                "pdbutility", pdb_name.replace(".", "_")
            )

        return cls.load_windows_symbol_table(
            context,
            guid,
            age,
            pdb_name,
            symbol_table_class,
            config_path,
            progress_callback,
        )

    @classmethod
    def load_windows_symbol_table(
        cls,
        context: interfaces.context.ContextInterface,
        guid: str,
        age: int,
        pdb_name: str,
        symbol_table_class: str,
        config_path: str = "pdbutility",
        progress_callback: constants.ProgressCallback = None,
    ):
        """Loads (downloading if necessary) a windows symbol table"""

        filter_string = os.path.join(
            pdb_name.strip("\x00"), guid.upper() + "-" + str(age)
        )

        isf_path = None
        # Take the first result of search for the intermediate file
        if not requirements.VersionRequirement.matches_required(
            (1, 0, 0), symbol_cache.SqliteCache.version
        ):
            vollog.debug("Required version of SQLiteCache not found")
            return None

        identifiers_path = os.path.join(
            constants.CACHE_PATH, constants.IDENTIFIERS_FILENAME
        )
        value = symbol_cache.SqliteCache(identifiers_path).find_location(
            symbol_cache.WindowsIdentifier.generate(
                pdb_name.strip("\x00"), guid.upper(), age
            ),
            "windows",
        )

        if value:
            isf_path = value
        else:
            # If none are found, attempt to download the pdb, convert it and try again
            cls.download_pdb_isf(
                context, guid.upper(), age, pdb_name, progress_callback
            )
            # Try again
            for value in intermed.IntermediateSymbolTable.file_symbol_url(
                "windows", filter_string
            ):
                isf_path = value
                break

        if not isf_path:
            vollog.debug(f"Required symbol library path not found: {filter_string}")
            vollog.info(
                "The symbols can be downloaded later using pdbconv.py -p {} -g {}".format(
                    pdb_name.strip("\x00"), guid.upper() + str(age)
                )
            )
            return None

        vollog.debug(f"Using symbol library: {filter_string}")

        # Set the discovered options
        join = interfaces.configuration.path_join
        context.config[join(config_path, "class")] = symbol_table_class
        context.config[join(config_path, "isf_url")] = isf_path
        parent_config_path = interfaces.configuration.parent_path(config_path)
        requirement_name = interfaces.configuration.path_head(config_path)

        # Construct the appropriate symbol table
        requirement = requirements.SymbolTableRequirement(
            name=requirement_name, description="PDBUtility generated symbol table"
        )
        requirement.construct(context, parent_config_path)
        return context.config[config_path]

    @classmethod
    def get_guid_from_mz(
        cls, context: interfaces.context.ContextInterface, layer_name: str, offset: int
    ) -> Optional[Tuple[str, int, str]]:
        """Takes the offset to an MZ header, locates any available pdb headers, and extracts the guid, age and pdb_name from them

        Args:
            context: The context on which to operate
            layer_name: The name of the (contiguous) layer within the context that contains the MZ file
            offset: The offset in the layer at which the MZ file begins

        Returns:
            A tuple of the guid, age and pdb_name, or None if no PDB record can be found
        """
        try:
            import pefile
        except ImportError:
            vollog.error(
                "Get_guid_from_mz requires the following python module: pefile"
            )
            return None

        layer = context.layers[layer_name]
        mz_sig = layer.read(offset, 2)

        # Check it is actually the MZ header
        if mz_sig != b"MZ":
            return None

        (nt_header_start,) = struct.unpack("<I", layer.read(offset + 0x3C, 4))
        pe_sig = layer.read(offset + nt_header_start, 2)

        # Check it is actually the Nt Headers
        if pe_sig != b"PE":
            return None

        (optional_header_size,) = struct.unpack(
            "<H", layer.read(offset + nt_header_start + 0x14, 2)
        )
        # Just enough to tell us the max size
        pe_header = layer.read(offset, nt_header_start + 0x16 + optional_header_size)
        pe_data = pefile.PE(data=pe_header)
        max_size = pe_data.OPTIONAL_HEADER.SizeOfImage

        # Proper data
        virtual_data = layer.read(offset, max_size, pad=True)
        pe_data = pefile.PE(data=virtual_data)

        # De-virtualize the memory
        sizeofHdrs = pe_data.OPTIONAL_HEADER.SizeOfHeaders
        physical_data = virtual_data[:sizeofHdrs]
        # Might need to put them in order by PointerToRawData just validate they are in order
        for sect in pe_data.sections:
            physical_data += virtual_data[
                sect.VirtualAddress : sect.VirtualAddress + sect.SizeOfRawData
            ]

        pe_data = pefile.PE(data=physical_data)

        if not hasattr(pe_data, "DIRECTORY_ENTRY_DEBUG") or not len(
            pe_data.DIRECTORY_ENTRY_DEBUG
        ):
            return None

        # Swap the Pointer with the Address since the de-virtualization doesn't apply to the fields
        debug_entry = None
        for index in range(len(pe_data.DIRECTORY_ENTRY_DEBUG)):
            if pe_data.DIRECTORY_ENTRY_DEBUG[index].struct.Type == 2:
                debug_data = pe_data.DIRECTORY_ENTRY_DEBUG[index]
                pe_data.set_dword_at_offset(
                    debug_data.struct.get_field_absolute_offset("AddressOfRawData"),
                    debug_data.struct.PointerToRawData,
                )
                pe_data.full_load()
                debug_entry = pe_data.DIRECTORY_ENTRY_DEBUG[index].entry

        if debug_entry is None:
            return None

        pdb_name = debug_entry.PdbFileName.decode("utf-8").strip("\x00")

        # Let pathlib do the filename extraction. This will likely always be a Windows path though.
        pdb_name = PureWindowsPath(pdb_name).name

        age = debug_entry.Age
        guid = debug_entry.Signature_String[:32]  # Removes the Age from the GUID
        return guid, age, pdb_name

    @classmethod
    def download_pdb_isf(
        cls,
        context: interfaces.context.ContextInterface,
        guid: str,
        age: int,
        pdb_name: str,
        progress_callback: constants.ProgressCallback = None,
    ) -> None:
        """Attempts to download the PDB file, convert it to an ISF file and
        save it to one of the symbol locations."""
        # Check for writability
        filter_string = os.path.join(pdb_name, guid + "-" + str(age))
        for path in symbols.__path__:
            # Store any temporary files created by downloading PDB files
            tmp_files = []
            potential_output_filename = os.path.join(
                path, "windows", filter_string + ".json.xz"
            )
            data_written = False
            try:
                os.makedirs(os.path.dirname(potential_output_filename), exist_ok=True)
                with lzma.open(potential_output_filename, "w") as of:
                    # Once we haven't thrown an error, do the computation
                    filename = pdbconv.PdbRetreiver().retreive_pdb(
                        guid + str(age),
                        file_name=pdb_name,
                        progress_callback=progress_callback,
                    )
                    if filename:
                        url = parse.urlparse(filename, scheme="file")
                        if url.scheme == "file" or len(url.scheme) == 1:
                            tmp_files.append(filename)
                            location = "file:" + request.pathname2url(
                                os.path.abspath(tmp_files[-1])
                            )
                        else:
                            location = filename
                        json_output = pdbconv.PdbReader(
                            context, location, pdb_name, progress_callback
                        ).get_json()
                        of.write(
                            bytes(
                                json.dumps(json_output, indent=2, sort_keys=True),
                                "utf-8",
                            )
                        )
                        # After we've successfully written it out, record the fact so we don't clear it out
                        data_written = True
                    else:
                        vollog.warning(
                            "Symbol file could not be downloaded from remote server"
                            + (" " * 100)
                        )
                break
            except PermissionError:
                vollog.debug(
                    f"Cannot write necessary symbol file, please check permissions on {potential_output_filename}"
                )
                continue
            finally:
                # If something else failed, removed the symbol file so we don't pick it up in the future
                if not data_written and os.path.exists(potential_output_filename):
                    os.remove(potential_output_filename)
                # Clear out all the temporary file if we constructed one
                for filename in tmp_files:
                    try:
                        os.remove(filename)
                    except PermissionError:
                        vollog.warning(
                            f"Temporary file could not be removed: {filename}"
                        )
        else:
            vollog.warning(
                "Cannot write downloaded symbols, please add the appropriate symbols"
                " or add/modify a symbols directory that is writable"
            )

    @classmethod
    def pdbname_scan(
        cls,
        ctx: interfaces.context.ContextInterface,
        layer_name: str,
        page_size: int,
        pdb_names: List[bytes],
        progress_callback: constants.ProgressCallback = None,
        start: Optional[int] = None,
        end: Optional[int] = None,
        maximum_invalid_count: int = 100,
    ) -> Generator[Dict[str, Optional[Union[bytes, str, int]]], None, None]:
        """Scans through `layer_name` at `ctx` looking for RSDS headers that
        indicate one of four common pdb kernel names (as listed in
        `self.pdb_names`) and returns the tuple (GUID, age, pdb_name,
        signature_offset, mz_offset)

        .. note:: This is automagical and therefore not guaranteed to provide correct results.

        The UI should always provide the user an opportunity to specify the
        appropriate types and PDB values themselves
        Args:
            layer_name: The layer name to scan
            page_size: Size of page constant
            pdb_names: List of pdb names to scan
            progress_callback: Means of providing the user with feedback during long processes
            start: Start address to start scanning from the pdb_names
            end: Minimum address to scan the pdb_names
            maximum_invalid_count: Amount of pages that can be invalid during scanning before aborting signature search
        """
        min_pfn = 0

        if start is None:
            start = ctx.layers[layer_name].minimum_address
        if end is None:
            end = ctx.layers[layer_name].maximum_address

        for GUID, age, pdb_name, signature_offset in ctx.layers[layer_name].scan(
            ctx,
            PdbSignatureScanner(pdb_names),
            progress_callback=progress_callback,
            sections=[(start, end - start)],
        ):
            mz_offset = None
            sig_pfn = signature_offset // page_size
            current_invalid_counter = 0

            for i in range(sig_pfn, min_pfn, -1):
                if current_invalid_counter > maximum_invalid_count:
                    break

                if not ctx.layers[layer_name].is_valid(i * page_size, 2):
                    current_invalid_counter += 1
                    continue

                data = ctx.layers[layer_name].read(i * page_size, 2)
                if data == b"MZ":
                    mz_offset = i * page_size
                    break
            min_pfn = sig_pfn

            yield {
                "GUID": GUID,
                "age": age,
                "pdb_name": str(pdb_name, "utf-8"),
                "signature_offset": signature_offset,
                "mz_offset": mz_offset,
            }

    @classmethod
    def symbol_table_from_pdb(
        cls,
        context: interfaces.context.ContextInterface,
        config_path: str,
        layer_name: str,
        pdb_name: str,
        module_offset: Optional[int] = None,
        module_size: Optional[int] = None,
    ) -> str:
        """Creates symbol table for a module in the specified layer_name.

        Searches the memory section of the loaded module for its PDB GUID
        and loads the associated symbol table into the symbol space.

        Args:
            context: The context to retrieve required elements (layers, symbol tables) from
            config_path: The config path where to find symbol files
            layer_name: The name of the layer on which to operate
            module_offset: This memory dump's module image offset
            module_size: The size of the module for this dump

        Returns:
            The name of the constructed and loaded symbol table
        """
        _, symbol_table_name = cls._modtable_from_pdb(
            context, config_path, layer_name, pdb_name, module_offset, module_size
        )

        if symbol_table_name is None:
            raise exceptions.SymbolSpaceError(
                f"Symbol table could not be reconstructed for module {pdb_name}"
            )

        return symbol_table_name

    @classmethod
    def _modtable_from_pdb(
        cls,
        context: interfaces.context.ContextInterface,
        config_path: str,
        layer_name: str,
        pdb_name: str,
        module_offset: Optional[int] = None,
        module_size: Optional[int] = None,
        create_module: bool = False,
    ) -> Tuple[Optional[str], Optional[str]]:
        if module_offset is None:
            module_offset = context.layers[layer_name].minimum_address
        if module_size is None:
            module_size = context.layers[layer_name].maximum_address - module_offset

        guids = list(
            cls.pdbname_scan(
                context,
                layer_name,
                context.layers[layer_name].page_size,
                [bytes(pdb_name, "latin-1")],
                start=module_offset,
                end=module_offset + module_size,
            )
        )

        if not guids:
            raise exceptions.SymbolSpaceError(
                f"Did not find GUID of {pdb_name} in module @ 0x{module_offset:x}!"
            )

        guid = guids[0]

        vollog.debug(f"Found {guid['pdb_name']}: {guid['GUID']}-{guid['age']}")

        module_name = guid["pdb_name"].replace(".pdb", "")

        symbol_table_name = cls.load_windows_symbol_table(
            context,
            guid["GUID"],
            guid["age"],
            guid["pdb_name"],
            "volatility3.framework.symbols.intermed.IntermediateSymbolTable",
            config_path=config_path,
        )

        new_module_name = None
        if create_module:
            new_module = contexts.Module.create(
                context,
                module_name,
                layer_name,
                offset=guid["mz_offset"],
                symbol_table_name=symbol_table_name,
            )
            new_module_name = new_module.name

        return new_module_name, symbol_table_name

    @classmethod
    def module_from_pdb(
        cls,
        context: interfaces.context.ContextInterface,
        config_path: str,
        layer_name: str,
        pdb_name: str,
        module_offset: Optional[int] = None,
        module_size: Optional[int] = None,
    ) -> str:
        """Creates a module in the specified layer_name based on a pdb name.

        Searches the memory section of the loaded module for its PDB GUID
        and loads the associated symbol table into the symbol space.

        Args:
            context: The context to retrieve required elements (layers, symbol tables) from
            config_path: The config path where to find symbol files
            layer_name: The name of the layer on which to operate
            module_offset: This memory dump's module image offset
            module_size: The size of the module for this dump

        Returns:
            The name of the constructed and loaded symbol table
        """

        module_name, _ = cls._modtable_from_pdb(
            context,
            config_path,
            layer_name,
            pdb_name,
            module_offset,
            module_size,
            create_module=True,
        )

        return module_name


class PdbSignatureScanner(interfaces.layers.ScannerInterface):
    """A :class:`~volatility3.framework.interfaces.layers.ScannerInterface`
    based scanner use to identify Windows PDB records.

    Args:
        pdb_names: A list of bytestrings, used to match pdb signatures against the pdb names within the records.

    .. note:: The pdb_names must be a list of byte strings, unicode strs will not match against the data scanned
    """

    _version = (1, 0, 0)

    _required_framework_version = (2, 27, 0)

    overlap = 0x4000
    """The size of overlap needed for the signature to ensure data cannot hide between two scanned chunks"""
    thread_safe = True
    """Determines whether the scanner accesses global variables in a thread safe manner (for use with :mod:`multiprocessing`)"""

    _RSDS_format = struct.Struct("<16BI")

    def __init__(self, pdb_names: List[bytes]) -> None:
        super().__init__()
        self._pdb_names = pdb_names

    def __call__(
        self, data: bytes, data_offset: int
    ) -> Generator[Tuple[str, Any, bytes, int], None, None]:
        pattern = (
            b"RSDS"
            + (b"." * self._RSDS_format.size)
            + b"("
            + b"|".join([re.escape(x) for x in self._pdb_names])
            + b")\x00"
        )
        for match in re.finditer(pattern, data, flags=re.DOTALL):
            pdb_name = data[
                match.start(0) + 4 + self._RSDS_format.size : match.start(0)
                + len(match.group())
                - 1
            ]
            if pdb_name in self._pdb_names:
                ## this ordering is intentional due to mixed endianness in the GUID
                (
                    g3,
                    g2,
                    g1,
                    g0,
                    g5,
                    g4,
                    g7,
                    g6,
                    g8,
                    g9,
                    ga,
                    gb,
                    gc,
                    gd,
                    ge,
                    gf,
                    a,
                ) = self._RSDS_format.unpack(
                    data[
                        match.start(0) + 4 : match.start(0) + 4 + self._RSDS_format.size
                    ]
                )

                guid = (16 * "{:02X}").format(
                    g0, g1, g2, g3, g4, g5, g6, g7, g8, g9, ga, gb, gc, gd, ge, gf
                )
                if match.start(0) < self.chunk_size:
                    yield (guid, a, pdb_name, data_offset + match.start(0))



# This file is Copyright 2024 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#

import logging
from typing import List

from volatility3.framework import constants, exceptions, interfaces, renderers
from volatility3.framework.configuration import requirements
from volatility3.framework.renderers import format_hints
from volatility3.plugins.windows import ssdt
from volatility3.plugins.windows.registry import hivelist

vollog = logging.getLogger(__name__)


class GetCellRoutine(interfaces.plugins.PluginInterface):
    """Reports registry hives with a hooked GetCellRoutine handler"""

    _required_framework_version = (2, 0, 0)

    @classmethod
    def get_requirements(cls) -> List[interfaces.configuration.RequirementInterface]:
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Windows kernel",
                architectures=["Intel32", "Intel64"],
            ),
            requirements.VersionRequirement(
                name="hivelist", component=hivelist.HiveList, version=(2, 0, 0)
            ),
            requirements.VersionRequirement(
                name="ssdt", component=ssdt.SSDT, version=(2, 0, 0)
            ),
        ]

    def _generator(self):
        collection = ssdt.SSDT.build_module_collection(
            context=self.context, kernel_module_name=self.config["kernel"]
        )

        # walk each hive and validate that the GetCellRoutine handler
        # is inside of the kernel (ntoskrnl)
        for hive_object in hivelist.HiveList.list_hives(
            context=self.context,
            base_config_path=self.config_path,
            kernel_module_name=self.config["kernel"],
        ):
            hive = hive_object.hive

            try:
                cellroutine = hive.GetCellRoutine
            except exceptions.InvalidAddressException:
                continue

            module_symbols = list(
                collection.get_module_symbols_by_absolute_location(cellroutine)
            )

            if module_symbols:
                for module_name, _ in module_symbols:
                    # GetCellRoutine handlers should only be in the kernel
                    if module_name not in constants.windows.KERNEL_MODULE_NAMES:
                        yield (
                            0,
                            (
                                format_hints.Hex(hive.vol.offset),
                                hive_object.get_name() or "",
                                module_name,
                                format_hints.Hex(cellroutine),
                            ),
                        )
            # Doesn't map to any module...
            else:
                yield (
                    0,
                    (
                        format_hints.Hex(hive.vol.offset),
                        hive_object.get_name() or "",
                        renderers.NotAvailableValue(),
                        format_hints.Hex(cellroutine),
                    ),
                )

    def run(self):
        return renderers.TreeGrid(
            [
                ("Hive Offset", renderers.format_hints.Hex),
                ("Hive Name", str),
                ("GetCellRoutine Module", str),
                ("GetCellRoutine Handler", renderers.format_hints.Hex),
            ],
            self._generator(),
        )



# This file is Copyright 2025 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#
import logging

from volatility3.framework import exceptions, interfaces, renderers
from volatility3.framework.configuration import requirements
from volatility3.framework.objects import utility
from volatility3.framework.renderers import format_hints
from volatility3.plugins.windows import pslist, pe_symbols

vollog = logging.getLogger(__name__)


# EtwpEventWriteFull -> https://github.com/SolitudePy/Stealthy-ETW-Patch
# CAPA rule -> https://github.com/mandiant/capa-rules/blob/master/anti-analysis/anti-av/patch-event-tracing-for-windows-function.yml
class EtwPatch(interfaces.plugins.PluginInterface):
    """Identifies ETW (Event Tracing for Windows) patching techniques used by malware to evade detection.

    This plugin examines the first opcode of key ETW functions in ntdll.dll and advapi32.dll
    to detect common ETW bypass techniques such as return pointer manipulation (RET) or function
    redirection (JMP). Attackers often patch these functions to prevent security tools from
    receiving telemetry about process execution, API calls, and other system events.
    """

    _version = (1, 0, 0)
    _required_framework_version = (2, 26, 0)

    etw_functions = {
        "ntdll.dll": {
            pe_symbols.wanted_names_identifier: [
                "EtwEventWrite",
                "EtwEventWriteFull",
                "NtTraceEvent",
                "ZwTraceEvent",
                "NtTraceControl",
                "ZwTraceControl",
                "EtwpEventWriteFull",
            ],
        },
        "advapi32.dll": {
            pe_symbols.wanted_names_identifier: ["EventWrite", "TraceEvent"],
        },
    }

    @classmethod
    def get_requirements(cls):
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Windows kernel",
                architectures=["Intel32", "Intel64"],
            ),
            requirements.VersionRequirement(
                name="pslist", component=pslist.PsList, version=(3, 0, 0)
            ),
            requirements.VersionRequirement(
                name="pe_symbols", component=pe_symbols.PESymbols, version=(3, 0, 0)
            ),
            requirements.ListRequirement(
                name="pid",
                description="Filter on specific process IDs",
                element_type=int,
                optional=True,
            ),
        ]

    def _generator(self):
        # Get all ETW function addresses before looping through processes
        found_symbols = pe_symbols.PESymbols.addresses_for_process_symbols(
            context=self.context,
            config_path=self.config_path,
            kernel_module_name=self.config["kernel"],
            symbols=self.etw_functions,
        )

        filter_func = pslist.PsList.create_pid_filter(self.config.get("pid", None))

        for proc in pslist.PsList.list_processes(
            context=self.context,
            kernel_module_name=self.config["kernel"],
            filter_func=filter_func,
        ):
            try:
                proc_id = proc.UniqueProcessId
                proc_name = utility.array_to_string(proc.ImageFileName)
                proc_layer_name = proc.add_process_layer()
            except exceptions.InvalidAddressException:
                vollog.debug(f"Unable to create process layer for PID {proc_id}")
                continue

            # Map of opcodes to their instruction names
            opcode_map = {
                0xC3: "RET",
                0xE9: "JMP",
            }

            for dll_name, functions in found_symbols.items():
                for func_name, func_addr in functions:
                    try:
                        opcode = self.context.layers[proc_layer_name].read(
                            func_addr, 1
                        )[0]
                        if opcode in opcode_map:
                            instruction = opcode_map[opcode]
                            yield (
                                0,
                                (
                                    proc_id,
                                    proc_name,
                                    dll_name,
                                    func_name,
                                    format_hints.Hex(func_addr),
                                    f"{opcode:02x} ({instruction})",
                                ),
                            )
                    except exceptions.InvalidAddressException:
                        vollog.debug(
                            f"Invalid address when reading function {func_name} at {func_addr:#x} in process {proc_id}"
                        )

    def run(self):
        return renderers.TreeGrid(
            [
                ("PID", int),
                ("Process", str),
                ("DLL", str),
                ("Function", str),
                ("Offset", format_hints.Hex),
                ("Opcode", str),
            ],
            self._generator(),
        )



# This file is Copyright 2024 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#

import logging

from typing import Iterator, List, Tuple, Iterable

from volatility3.framework import (
    renderers,
    interfaces,
    constants,
    symbols,
    exceptions,
)
from volatility3.framework.configuration import requirements
from volatility3.framework.renderers import format_hints
from volatility3.framework.symbols.windows import versions, extensions
from volatility3.plugins.windows import ssdt, kpcrs

vollog = logging.getLogger(__name__)


class Timers(interfaces.plugins.PluginInterface):
    """Print kernel timers and associated module DPCs"""

    _required_framework_version = (2, 0, 0)
    _version = (1, 0, 1)

    @classmethod
    def get_requirements(cls) -> List[interfaces.configuration.RequirementInterface]:
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Windows kernel",
                architectures=["Intel32", "Intel64"],
            ),
            requirements.VersionRequirement(
                name="ssdt", component=ssdt.SSDT, version=(2, 0, 0)
            ),
            requirements.VersionRequirement(
                name="kpcrs", component=kpcrs.KPCRs, version=(2, 0, 0)
            ),
        ]

    @classmethod
    def list_timers(
        cls,
        context: interfaces.context.ContextInterface,
        kernel_module_name: str,
    ) -> Iterable[extensions.KTIMER]:
        """Lists all kernel timers.

        Args:
            context: The context to retrieve required elements (layers, symbol tables) from
            kernel_module_name: The name of the kernel module on which to operate

        Yields:
            A _KTIMER entry
        """

        kernel = context.modules[kernel_module_name]
        if versions.is_windows_7(
            context=context, symbol_table=kernel.symbol_table_name
        ) or versions.is_windows_8_or_later(
            context=context, symbol_table=kernel.symbol_table_name
        ):
            # Starting with Windows 7, there is no more KiTimerTableListHead. The list is
            # at _KPCR.PrcbData.TimerTable.TimerEntries
            # See http://pastebin.com/FiRsGW3f
            for kpcr, _ in kpcrs.KPCRs.list_kpcrs(context, kernel_module_name):
                if hasattr(kpcr.Prcb.TimerTable, "TableState"):
                    for timer_entries in kpcr.Prcb.TimerTable.TimerEntries:
                        for timer_entry in timer_entries:
                            for timer in timer_entry.Entry.to_list(
                                kernel.symbol_table_name + constants.BANG + "_KTIMER",
                                "TimerListEntry",
                            ):
                                yield timer

                else:
                    for timer_entries in kpcr.Prcb.TimerTable.TimerEntries:
                        for timer in timer_entries.Entry.to_list(
                            kernel.symbol_table_name + constants.BANG + "_KTIMER",
                            "TimerListEntry",
                        ):
                            yield timer

        elif versions.is_xp_or_2003(
            context=context, symbol_table=kernel.symbol_table_name
        ) or versions.is_vista_or_later(
            context=context, symbol_table=kernel.symbol_table_name
        ):
            is_64bit = symbols.symbol_table_is_64bit(context, kernel.symbol_table_name)
            if is_64bit or versions.is_vista_or_later(
                context=context, symbol_table=kernel.symbol_table_name
            ):
                # On XP x64, Windows 2003 SP1-SP2, and Vista SP0-SP2, KiTimerTableListHead
                # is an array of 512 _KTIMER_TABLE_ENTRY structs.
                array_size = 512
            else:
                # On XP SP0-SP3 x86 and Windows 2003 SP0, KiTimerTableListHead
                # is an array of 256 _LIST_ENTRY for _KTIMERs.
                array_size = 256

            timer_table_list_head = kernel.object(
                object_type="array",
                offset=kernel.get_symbol("KiTimerTableListHead").address,
                subtype=kernel.get_type("_LIST_ENTRY"),
                count=array_size,
            )
            for table in timer_table_list_head:
                for timer in table.to_list(
                    kernel.symbol_table_name + constants.BANG + "_KTIMER",
                    "TimerListEntry",
                ):
                    yield timer

        else:
            raise NotImplementedError("This version of Windows is not supported!")

    def _generator(self) -> Iterator[Tuple]:
        collection = ssdt.SSDT.build_module_collection(
            context=self.context,
            kernel_module_name=self.config["kernel"],
        )

        # FIXME - the list_timers API is gross. Fix after GUI merge
        for timer in self.list_timers(
            self.context,
            self.config["kernel"],
        ):
            if not timer.valid_type():
                continue

            try:
                dpc = timer.get_dpc()
                if dpc == 0:
                    continue
                if dpc.DeferredRoutine == 0:
                    continue
                deferred_routine = dpc.DeferredRoutine
            except exceptions.InvalidAddressException as exc:
                vollog.debug(
                    f"Failed to get _KTIMER.Dpc due to {exc.__class__.__name__}"
                )
                continue

            module_symbols = list(
                collection.get_module_symbols_by_absolute_location(deferred_routine)
            )

            if module_symbols:
                for module_name, symbol_generator in module_symbols:
                    symbols_found = False

                    # we might have multiple symbols pointing to the same location
                    for symbol in symbol_generator:
                        symbols_found = True
                        yield (
                            0,
                            (
                                format_hints.Hex(timer.vol.offset),
                                timer.get_due_time(),
                                timer.Period,
                                timer.get_signaled(),
                                format_hints.Hex(deferred_routine),
                                module_name,
                                symbol.split(constants.BANG)[1],
                            ),
                        )

                    # no symbols, but we at least can report the module name
                    if not symbols_found:
                        yield (
                            0,
                            (
                                format_hints.Hex(timer.vol.offset),
                                timer.get_due_time(),
                                timer.Period,
                                timer.get_signaled(),
                                format_hints.Hex(deferred_routine),
                                module_name,
                                renderers.NotAvailableValue(),
                            ),
                        )
            else:
                # no module was found at the absolute location
                yield (
                    0,
                    (
                        format_hints.Hex(timer.vol.offset),
                        timer.get_due_time(),
                        timer.Period,
                        timer.get_signaled(),
                        format_hints.Hex(deferred_routine),
                        renderers.NotAvailableValue(),
                        renderers.NotAvailableValue(),
                    ),
                )

    def run(self):
        return renderers.TreeGrid(
            [
                ("Offset", format_hints.Hex),
                ("DueTime", str),
                ("Period(ms)", int),
                ("Signaled", str),
                ("Routine", format_hints.Hex),
                ("Module", str),
                ("Symbol", str),
            ],
            self._generator(),
        )



# This file is Copyright 2024 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#
import logging

from typing import Optional, Tuple, Generator, Dict

from volatility3.framework import interfaces, exceptions
from volatility3.framework import renderers
from volatility3.framework.configuration import requirements
from volatility3.framework.objects import utility
from volatility3.framework.renderers import format_hints
from volatility3.plugins.windows import pslist, vadinfo

vollog = logging.getLogger(__name__)


class ProcessGhosting(interfaces.plugins.PluginInterface):
    """Lists processes whose DeletePending bit is set or whose FILE_OBJECT is set to 0 or Vads that are DeleteOnClose"""

    _version = (1, 0, 0)
    _required_framework_version = (2, 4, 0)

    @classmethod
    def get_requirements(cls):
        # Since we're calling the plugin, make sure we have the plugin's requirements
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Windows kernel",
                architectures=["Intel32", "Intel64"],
            ),
            requirements.VersionRequirement(
                name="pslist", component=pslist.PsList, version=(3, 0, 0)
            ),
            requirements.VersionRequirement(
                name="vadinfo", component=vadinfo.VadInfo, version=(2, 0, 1)
            ),
        ]

    @classmethod
    def _process_checks(
        cls,
        proc: interfaces.objects.ObjectInterface,
        mapped_files: Dict[int, Tuple[str, interfaces.objects.ObjectInterface]],
    ) -> Generator[
        Tuple[int, Optional[int], Optional[int], int, Optional[str]], None, None
    ]:
        """
        Checks the EPROCESS for signs of ghosting
        """
        if not proc.has_member("ImageFilePointer"):
            return

        delete_pending = None

        # if it is 0 then its a side effect of process ghosting
        if proc.ImageFilePointer.vol.offset != 0:
            try:
                file_object = proc.ImageFilePointer
                delete_pending = file_object.DeletePending
                file_object = file_object.dereference().vol.offset
            except exceptions.InvalidAddressException:
                file_object = 0

        # ImageFilePointer equal to 0 means process ghosting or similar techniques were used
        else:
            file_object = 0

        # delete_pending besides 0 or 1 = smear
        if isinstance(delete_pending, int) and delete_pending not in [0, 1]:
            vollog.debug(
                f"Invalid delete_pending value {delete_pending} found for process {proc.UniqueProcessId}"
            )
            delete_pending = None

        if file_object == 0 or delete_pending == 1:
            yield file_object, delete_pending, None, proc.SectionBaseAddress

    @classmethod
    def _vad_checks(
        cls, control_area: interfaces.objects.ObjectInterface, vad_path: str
    ) -> Generator[Tuple[int, Optional[int], Optional[int]], None, None]:
        """
        Checks the control area for delete on close or delete pending being set
        """
        try:
            file_object = control_area.FilePointer.dereference().cast("_FILE_OBJECT")
        except exceptions.InvalidAddressException:
            return

        try:
            delete_on_close = control_area.u.Flags.DeleteOnClose
        except exceptions.InvalidAddressException:
            delete_on_close = None

        if delete_on_close and vad_path.lower().endswith((".exe", ".dll")):
            yield file_object.vol.offset, None, delete_on_close

        try:
            delete_pending = file_object.DeletePending
        except exceptions.InvalidAddressException:
            delete_pending = None

        if delete_pending == 1:
            yield file_object.vol.offset, delete_pending, None

    @classmethod
    def check_for_ghosting(
        cls,
        proc: interfaces.objects.ObjectInterface,
        mapped_files: Dict[int, Tuple[str, interfaces.objects.ObjectInterface]],
    ) -> Generator[
        Tuple[int, Optional[int], Optional[int], int, Optional[str]], None, None
    ]:
        """
        Returns process or vad info for ghosting files

        Args:
            proc:
            mapped_files: A dictionary mapping vad base addresses to the path and vad instance for the process

        Return:
            A Generator of tuples of the file object address, the delete pending state, delete on close state, base address of the VAD, and the path
        """
        # check the direct file object of the process
        yield from cls._process_checks(proc, mapped_files)

        # walk each vad, check if it is pending delete or has its delete on close bit set
        for vad_base, (path, vad) in mapped_files.items():
            # these checks have no meaning for private memory areas
            if vad.get_private_memory() == 1:
                continue

            try:
                if vad.has_member("ControlArea"):
                    control_area = vad.ControlArea
                elif vad.has_member("Subsection"):
                    control_area = vad.Subsection.ControlArea
                # We got here from a short vad, likely smear
                else:
                    continue
            except exceptions.InvalidAddressException:
                vollog.debug(
                    f"Unable to get control area for vad at base {vad_base:#x} for process with pid {proc.UniqueProcessId}"
                )
                continue

            for file_object_address, delete_pending, delete_on_close in cls._vad_checks(
                control_area, path
            ):
                yield (
                    format_hints.Hex(file_object_address),
                    delete_pending,
                    delete_on_close,
                    vad_base,
                )

    def _generator(self, procs):
        kernel = self.context.modules[self.config["kernel"]]

        has_imagefilepointer = kernel.get_type("_EPROCESS").has_member(
            "ImageFilePointer"
        )
        if not has_imagefilepointer:
            vollog.warning(
                "ImageFilePointer checks are only supported on Windows 10+ builds when the ImageFilePointer member of _EPROCESS is present"
            )

        for proc in procs:
            process_name = utility.array_to_string(proc.ImageFileName)
            pid = proc.UniqueProcessId

            # base address -> (file path, VAD instance)
            mapped_files: Dict[int, Tuple[str, interfaces.objects.ObjectInterface]] = {}
            for vad in vadinfo.VadInfo.list_vads(proc):
                path = vad.get_file_name()
                if isinstance(path, str):
                    mapped_files[vad.get_start()] = (path, vad)

            for (
                file_object_address,
                delete_pending,
                delete_on_close,
                base_address,
            ) in self.check_for_ghosting(proc, mapped_files):
                vad_info = mapped_files.get(base_address)
                if vad_info:
                    path = vad_info[0]
                else:
                    path = renderers.NotAvailableValue()

                yield (
                    0,
                    (
                        pid,
                        process_name,
                        format_hints.Hex(base_address),
                        format_hints.Hex(file_object_address),
                        delete_pending or renderers.NotApplicableValue(),
                        delete_on_close or renderers.NotApplicableValue(),
                        path,
                    ),
                )

    def run(self):
        filter_func = pslist.PsList.create_active_process_filter()

        return renderers.TreeGrid(
            [
                ("PID", int),
                ("Process", str),
                ("Base", format_hints.Hex),
                ("FILE_OBJECT", format_hints.Hex),
                ("DeletePending", int),
                ("DeleteOnClose", int),
                ("Path", str),
            ],
            self._generator(
                pslist.PsList.list_processes(
                    context=self.context,
                    kernel_module_name=self.config["kernel"],
                    filter_func=filter_func,
                )
            ),
        )



# This file is Copyright 2019 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#
import logging
from typing import Iterator, List, Optional, Tuple

from volatility3.framework import exceptions, interfaces, renderers
from volatility3.framework.configuration import requirements
from volatility3.framework.layers import registry
from volatility3.framework.renderers import format_hints
from volatility3.plugins.windows.registry import hivescan

vollog = logging.getLogger(__name__)


class HiveGenerator:
    """Walks the registry HiveList linked list in a given direction and stores an invalid offset
    if it's unable to fully walk the list"""

    _required_framework_version = (2, 0, 0)

    def __init__(self, cmhive, forward=True):
        self._cmhive = cmhive
        self._forward = forward
        self._invalid = None

    def __iter__(self):
        for hive in self._cmhive.HiveList.to_list(
            self._cmhive.vol.type_name, "HiveList", forward=self._forward
        ):
            if not hive.is_valid():
                self._invalid = hive.vol.offset
                return None
            yield hive

    @property
    def invalid(self) -> Optional[int]:
        return self._invalid


class HiveList(interfaces.plugins.PluginInterface):
    """Lists the registry hives present in a particular memory image."""

    _required_framework_version = (2, 0, 0)

    # 2.0.0 - changed the signature of list_hives
    _version = (2, 0, 0)

    @classmethod
    def get_requirements(cls) -> List[interfaces.configuration.RequirementInterface]:
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Windows kernel",
                architectures=["Intel32", "Intel64"],
            ),
            requirements.StringRequirement(
                name="filter",
                description="String to filter hive names returned",
                optional=True,
                default=None,
            ),
            requirements.VersionRequirement(
                name="hivescan", component=hivescan.HiveScan, version=(2, 0, 0)
            ),
            requirements.BooleanRequirement(
                name="dump",
                description="Extract listed registry hives",
                default=False,
                optional=True,
            ),
        ]

    def _sanitize_hive_name(self, name: str) -> str:
        return (
            name.split("\\")[-1]
            .replace(" ", "_")
            .replace(".", "")
            .replace("[", "")
            .replace("]", "")
        )

    def _generator(self) -> Iterator[Tuple[int, Tuple[int, str]]]:
        chunk_size = 0x500000
        kernel = self.context.modules[self.config["kernel"]]

        for hive_object in self.list_hive_objects(
            context=self.context,
            layer_name=kernel.layer_name,
            symbol_table=kernel.symbol_table_name,
            filter_string=self.config.get("filter", None),
        ):
            file_output = "Disabled"
            if self.config["dump"]:
                # Construct the hive
                hive = next(
                    self.list_hives(
                        context=self.context,
                        base_config_path=self.config_path,
                        kernel_module_name=self.config["kernel"],
                        hive_offsets=[hive_object.vol.offset],
                    )
                )
                maxaddr = hive.hive.Storage[0].Length
                hive_name = self._sanitize_hive_name(hive.get_name())

                file_handle = self.open(
                    f"registry.{hive_name}.{hex(hive.hive_offset)}.hive"
                )
                with file_handle as file_data:
                    if hive._base_block:
                        hive_data = self.context.layers[hive.dependencies[0]].read(
                            hive.hive.BaseBlock, 1 << 12
                        )
                    else:
                        hive_data = "\x00" * (1 << 12)
                    file_data.write(hive_data)

                    for i in range(0, maxaddr, chunk_size):
                        current_chunk_size = min(chunk_size, maxaddr - i)
                        data = hive.read(i, current_chunk_size, pad=True)
                        file_data.write(data)
                        # if self._progress_callback:
                        #     self._progress_callback((i / maxaddr) * 100, 'Writing layer {}'.format(hive_name))
                file_output = file_handle.preferred_filename

            yield (
                0,
                (
                    format_hints.Hex(hive_object.vol.offset),
                    hive_object.get_name() or "",
                    file_output,
                ),
            )

    @classmethod
    def list_hives(
        cls,
        context: interfaces.context.ContextInterface,
        base_config_path: str,
        kernel_module_name: str,
        filter_string: Optional[str] = None,
        hive_offsets: Optional[List[int]] = None,
    ) -> Iterator[registry.RegistryHive]:
        """Walks through a registry, hive by hive returning the constructed
        registry layer name.

        Args:
            context: The context to retrieve required elements (layers, symbol tables) from
            base_config_path: The configuration path for any settings required by the new table
            kernel_module_name: The name of the module for the kernel
            filter_string: An optional string which must be present in the hive name if specified
            offset: An optional offset to specify a specific hive to iterate over (takes precedence over filter_string)

        Yields:
            A registry hive layer name
        """
        kernel = context.modules[kernel_module_name]

        if hive_offsets is None:
            try:
                hive_offsets = [
                    hive.vol.offset
                    for hive in cls.list_hive_objects(
                        context=context,
                        layer_name=kernel.layer_name,
                        symbol_table=kernel.symbol_table_name,
                        filter_string=filter_string,
                    )
                ]
            except ImportError:
                vollog.warning(
                    "Unable to import windows.hivelist plugin, please provide a hive offset"
                )
                raise ValueError(
                    "Unable to import windows.hivelist plugin, please provide a hive offset"
                )

        for hive_offset in hive_offsets:
            # Construct the hive
            reg_config_path = cls.make_subconfig(
                context=context,
                base_config_path=base_config_path,
                hive_offset=hive_offset,
                base_layer=kernel.layer_name,
                nt_symbols=kernel.symbol_table_name,
                kernel_module_name=kernel_module_name,
            )

            try:
                hive = registry.RegistryHive(
                    context, reg_config_path, name="hive" + hex(hive_offset)
                )
            except exceptions.InvalidAddressException:
                vollog.warning(
                    f"Couldn't create RegistryHive layer at offset {hex(hive_offset)}, skipping"
                )
                continue
            context.layers.add_layer(hive)
            yield hive

    @classmethod
    def list_hive_objects(
        cls,
        context: interfaces.context.ContextInterface,
        layer_name: str,
        symbol_table: str,
        filter_string: Optional[str] = None,
    ) -> Iterator[interfaces.objects.ObjectInterface]:
        """Lists all the hives in the primary layer.

        Args:
            context: The context to retrieve required elements (layers, symbol tables) from
            layer_name: The name of the layer on which to operate
            symbol_table: The name of the table containing the kernel symbols
            filter_string: A string which must be present in the hive name if specified

        Returns:
            The list of registry hives from the `layer_name` layer as filtered against using the `filter_string`
        """

        # We only use the object factory to demonstrate how to use one
        kvo = context.layers[layer_name].config.get("kernel_virtual_offset", None)
        if not kvo:
            raise ValueError(
                "Intel layer does not have an associated kernel virtual offset, failing"
            )
        ntkrnlmp = context.module(symbol_table, layer_name=layer_name, offset=kvo)

        list_head = ntkrnlmp.get_symbol("CmpHiveListHead").address
        list_entry = ntkrnlmp.object(object_type="_LIST_ENTRY", offset=list_head)
        reloff = ntkrnlmp.get_type("_CMHIVE").relative_child_offset("HiveList")
        cmhive = ntkrnlmp.object(
            object_type="_CMHIVE", offset=list_entry.vol.offset - reloff, absolute=True
        )

        # Run through the list forwards
        seen = set()

        hg = HiveGenerator(cmhive, forward=True)
        for hive in hg:
            if hive.vol.offset in seen:
                vollog.debug(
                    f"Hivelist found an already seen offset {hex(hive.vol.offset)} while "
                    "traversing forwards, this should not occur"
                )
                break
            seen.add(hive.vol.offset)
            if (
                filter_string is None
                or filter_string.lower() in str(hive.get_name() or "").lower()
            ):
                if context.layers[layer_name].is_valid(hive.vol.offset):
                    yield hive

        forward_invalid = hg.invalid
        if forward_invalid:
            vollog.debug(
                f"Hivelist failed traversing the list forwards at {hex(forward_invalid)}, traversing backwards"
            )
            hg = HiveGenerator(cmhive, forward=False)
            for hive in hg:
                if hive.vol.offset in seen:
                    vollog.debug(
                        f"Hivelist found an already seen offset {hex(hive.vol.offset)} while "
                        "traversing backwards, list walking met in the middle"
                    )
                    break
                seen.add(hive.vol.offset)
                if (
                    filter_string is None
                    or filter_string.lower() in str(hive.get_name() or "").lower()
                ):
                    if context.layers[layer_name].is_valid(hive.vol.offset):
                        yield hive

            backward_invalid = hg.invalid

            if backward_invalid and forward_invalid != backward_invalid:
                # walking forward and backward did not stop at the same offset. they should if:
                #  1) there are no invalid hives, walking forwards would reach the end and backwards is not necessary
                #  2) there is one invalid hive, walking backwards would stop at the same place as forwards
                # therefore, there must be more 2 or more invalid hives, so the middle of the list is not reachable
                # by walking the list, so revert to scanning, and walk the list forwards and backwards from each
                # found hive
                vollog.debug(
                    f"Hivelist failed traversing backwards at {hex(backward_invalid)}, a different "
                    "location from forwards, revert to scanning"
                )
                for hive in hivescan.HiveScan.scan_hives(context, ntkrnlmp.name):
                    try:
                        if hive.HiveList.Flink:
                            start_hive_offset = hive.HiveList.Flink - reloff

                            ## Now instantiate the first hive in virtual address space as normal
                            start_hive = ntkrnlmp.object(
                                object_type="_CMHIVE",
                                offset=start_hive_offset,
                                absolute=True,
                            )
                            for forward in (True, False):
                                for linked_hive in start_hive.HiveList.to_list(
                                    hive.vol.type_name, "HiveList", forward
                                ):
                                    if (
                                        not linked_hive.is_valid()
                                        or linked_hive.vol.offset in seen
                                    ):
                                        continue
                                    seen.add(linked_hive.vol.offset)
                                    if (
                                        filter_string is None
                                        or filter_string.lower()
                                        in str(linked_hive.get_name() or "").lower()
                                    ):
                                        if context.layers[layer_name].is_valid(
                                            linked_hive.vol.offset
                                        ):
                                            yield linked_hive
                    except exceptions.InvalidAddressException:
                        vollog.debug(
                            f"InvalidAddressException when traversing hive {hex(hive.vol.offset)} found from scan, skipping"
                        )

    def run(self) -> renderers.TreeGrid:
        return renderers.TreeGrid(
            [("Offset", format_hints.Hex), ("FileFullPath", str), ("File output", str)],
            self._generator(),
        )



# This file is Copyright 2019 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#

import datetime
import logging
from typing import Callable, Iterator, List, Optional, Type

from volatility3.framework import constants, exceptions, interfaces, layers, renderers
from volatility3.framework.configuration import requirements
from volatility3.framework.objects import utility
from volatility3.framework.renderers import format_hints
from volatility3.framework.symbols import intermed
from volatility3.framework.symbols.windows import extensions
from volatility3.framework.symbols.windows.extensions import pe
from volatility3.plugins import timeliner

vollog = logging.getLogger(__name__)


class PsList(interfaces.plugins.PluginInterface, timeliner.TimeLinerInterface):
    """Lists the processes present in a particular windows memory image."""

    _required_framework_version = (2, 0, 0)

    # 3.0.0 - changed signature for `list_processes`
    _version = (3, 0, 1)
    PHYSICAL_DEFAULT = False

    @classmethod
    def get_requirements(cls):
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Windows kernel",
                architectures=["Intel32", "Intel64"],
            ),
            requirements.BooleanRequirement(
                name="physical",
                description="Display physical offsets instead of virtual",
                default=cls.PHYSICAL_DEFAULT,
                optional=True,
            ),
            requirements.VersionRequirement(
                name="timeliner",
                component=timeliner.TimeLinerInterface,
                version=(1, 0, 0),
            ),
            requirements.ListRequirement(
                name="pid",
                element_type=int,
                description="Process ID to include (all other processes are excluded)",
                optional=True,
            ),
            requirements.BooleanRequirement(
                name="dump",
                description="Extract listed processes",
                default=False,
                optional=True,
            ),
        ]

    @classmethod
    def process_dump(
        cls,
        context: interfaces.context.ContextInterface,
        kernel_table_name: str,
        pe_table_name: str,
        proc: interfaces.objects.ObjectInterface,
        open_method: Type[interfaces.plugins.FileHandlerInterface],
    ) -> interfaces.plugins.FileHandlerInterface:
        """Extracts the complete data for a process as a FileHandlerInterface

        Args:
            context: the context to operate upon
            kernel_table_name: the name for the symbol table containing the kernel's symbols
            pe_table_name: the name for the symbol table containing the PE format symbols
            proc: the process object whose memory should be output
            open_method: class to provide context manager for opening the file

        Returns:
            An open FileHandlerInterface object containing the complete data for the process or None in the case of failure
        """

        file_handle = None
        proc_id = "Invalid process object"
        try:
            proc_id = proc.UniqueProcessId
            proc_layer_name = proc.add_process_layer()
            peb = context.object(
                kernel_table_name + constants.BANG + "_PEB",
                layer_name=proc_layer_name,
                offset=proc.Peb,
            )

            dos_header = context.object(
                pe_table_name + constants.BANG + "_IMAGE_DOS_HEADER",
                offset=peb.ImageBaseAddress,
                layer_name=proc_layer_name,
            )

            process_name = proc.ImageFileName.cast(
                "string",
                max_length=proc.ImageFileName.vol.count,
                errors="replace",
            )

            file_handle = open_method(
                open_method.sanitize_filename(
                    f"{proc.UniqueProcessId}.{process_name}.{peb.ImageBaseAddress:#x}.dmp"
                )
            )

            for offset, data in dos_header.reconstruct():
                file_handle.seek(offset)
                file_handle.write(data)
        except Exception as excp:
            vollog.debug(f"Unable to dump PE with pid {proc_id}: {excp}")

        return file_handle

    @classmethod
    def create_pid_filter(
        cls, pid_list: Optional[List[int]] = None, exclude: bool = False
    ) -> Callable[[interfaces.objects.ObjectInterface], bool]:
        """A factory for producing filter functions that filter based on a list
        of process IDs.

        Args:
            pid_list: A list of process IDs that are acceptable, all other processes will be filtered out
            exclude: Accept only tasks that are not in pid_list

        Returns:
            Filter function for passing to the `list_processes` method
        """

        def filter_func(_):
            return False

        # FIXME: mypy #4973 or #2608
        pid_list = pid_list or []
        filter_list = [x for x in pid_list if x is not None]
        if filter_list:
            if exclude:

                def filter_func(x):
                    return x.UniqueProcessId in filter_list

            else:

                def filter_func(x):
                    return x.UniqueProcessId not in filter_list

        return filter_func

    @classmethod
    def create_active_process_filter(
        cls,
    ) -> Callable[[interfaces.objects.ObjectInterface], bool]:
        """A factory for producing a filter function that only returns
           active, userland processes. This prevents plugins from operating on terminated
           processes that are still in the process list due to smear or handle leaks as well
           as kernel processes (System, Registry, etc.). Use of this filter for plugins searching
           for system state anomalies significantly reduces false positive in smeared and terminated
           processes.
        Returns:
            Filter function for passing to the `list_processes` method
        """

        return lambda x: (
            not (
                x.is_valid()
                and x.ActiveThreads > 0
                and x.UniqueProcessId != 4
                and x.InheritedFromUniqueProcessId != 4
                and x.ExitTime.QuadPart == 0
                and x.get_handle_count() != renderers.UnreadableValue()
            )
        )

    @classmethod
    def create_name_filter(
        cls, name_list: Optional[List[str]] = None, exclude: bool = False
    ) -> Callable[[interfaces.objects.ObjectInterface], bool]:
        """A factory for producing filter functions that filter based on a list
        of process names.

        Args:
            name_list: A list of process names that are acceptable, all other processes will be filtered out
            exclude: Accept only tasks that are not in name_list
        Returns:
            Filter function for passing to the `list_processes` method
        """

        def filter_func(_):
            return False

        # FIXME: mypy #4973 or #2608
        name_list = name_list or []
        filter_list = [x for x in name_list if x is not None]
        if filter_list:
            if exclude:

                def filter_func(x):
                    return utility.array_to_string(x.ImageFileName) in filter_list

            else:

                def filter_func(x):
                    return utility.array_to_string(x.ImageFileName) not in filter_list

        return filter_func

    @classmethod
    def list_processes(
        cls,
        context: interfaces.context.ContextInterface,
        kernel_module_name: str,
        filter_func: Callable[[interfaces.objects.ObjectInterface], bool] = lambda _: (
            False
        ),
    ) -> Iterator["extensions.EPROCESS"]:
        """Lists all the processes in the given layer that are in the pid
        config option.

        Args:
            context: The context to retrieve required elements (layers, symbol tables) from
            layer_iname: The name of the layer on which to operate
            symbol_table_name: The name of the table containing the kernel symbols
            filter_func: A function which takes an EPROCESS object and returns True if the process should be ignored/filtered

        Returns:
            The list of EPROCESS objects from the `layer_name` layer's PsActiveProcessHead list after filtering
        """

        kernel = context.modules[kernel_module_name]

        if not kernel.offset:
            raise ValueError(
                "Intel layer does not have an associated kernel virtual offset, failing"
            )

        ps_aph_offset = kernel.get_symbol("PsActiveProcessHead").address
        list_entry = kernel.object(object_type="_LIST_ENTRY", offset=ps_aph_offset)

        # This is example code to demonstrate how to use symbol_space directly, rather than through a module:
        #
        # ```
        # reloff = self.context.symbol_space.get_type(
        #          self.config['nt_symbols'] + constants.BANG + "_EPROCESS").relative_child_offset(
        #          "ActiveProcessLinks")
        # ```
        #
        # Note: "nt_symbols!_EPROCESS" could have been used, but would rely on the "nt_symbols" symbol table not already
        # having been present.  Strictly, the value of the requirement should be joined with the BANG character
        # defined in the constants file
        reloff = kernel.get_type("_EPROCESS").relative_child_offset(
            "ActiveProcessLinks"
        )
        eproc = kernel.object(
            object_type="_EPROCESS",
            offset=list_entry.vol.offset - reloff,
            absolute=True,
        )

        seen = set()
        for forward in (True, False):
            for proc in eproc.ActiveProcessLinks.to_list(
                symbol_type=eproc.vol.type_name,
                member="ActiveProcessLinks",
                forward=forward,
            ):
                if proc.vol.offset in seen:
                    continue
                seen.add(proc.vol.offset)
                if not filter_func(proc):
                    yield proc

    def _generator(self):
        kernel = self.context.modules[self.config["kernel"]]

        pe_table_name = intermed.IntermediateSymbolTable.create(
            self.context, self.config_path, "windows", "pe", class_types=pe.class_types
        )

        memory = self.context.layers[kernel.layer_name]
        if not isinstance(memory, layers.intel.Intel):
            raise TypeError("Primary layer is not an intel layer")

        for proc in self.list_processes(
            self.context,
            self.config["kernel"],
            filter_func=self.create_pid_filter(self.config.get("pid", None)),
        ):
            if not self.config.get("physical", self.PHYSICAL_DEFAULT):
                offset = proc.vol.offset
            else:
                (_, _, offset, _, _) = list(
                    memory.mapping(offset=proc.vol.offset, length=0)
                )[0]

            file_output = "Disabled"

            try:
                if self.config["dump"]:
                    file_handle = self.process_dump(
                        self.context,
                        kernel.symbol_table_name,
                        pe_table_name,
                        proc,
                        self.open,
                    )
                    file_output = "Error outputting file"
                    if file_handle:
                        file_handle.close()
                        file_output = str(file_handle.preferred_filename)

                yield (
                    0,
                    (
                        proc.UniqueProcessId,
                        proc.InheritedFromUniqueProcessId,
                        proc.ImageFileName.cast(
                            "string",
                            max_length=proc.ImageFileName.vol.count,
                            errors="replace",
                        ),
                        format_hints.Hex(offset),
                        proc.ActiveThreads,
                        proc.get_handle_count(),
                        proc.get_session_id(),
                        proc.get_is_wow64(),
                        proc.get_create_time(),
                        proc.get_exit_time(),
                        file_output,
                    ),
                )

            except exceptions.InvalidAddressException:
                vollog.info(
                    f"Invalid process found at address: {proc.vol.offset:x}. Skipping"
                )

    def generate_timeline(self):
        for row in self._generator():
            _depth, row_data = row
            description = f"Process: {row_data[0]} {row_data[2]} ({row_data[3]})"
            yield (description, timeliner.TimeLinerType.CREATED, row_data[8])
            yield (description, timeliner.TimeLinerType.MODIFIED, row_data[9])

    def run(self):
        offsettype = (
            "(V)" if not self.config.get("physical", self.PHYSICAL_DEFAULT) else "(P)"
        )

        return renderers.TreeGrid(
            [
                ("PID", int),
                ("PPID", int),
                ("ImageFileName", str),
                (f"Offset{offsettype}", format_hints.Hex),
                ("Threads", int),
                ("Handles", int),
                ("SessionId", int),
                ("Wow64", bool),
                ("CreateTime", datetime.datetime),
                ("ExitTime", datetime.datetime),
                ("File output", str),
            ],
            self._generator(),
        )



# This file is Copyright 2019 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#
"""Renderers.

Renderers display the unified output format in some manner (be it text
or file or graphical output
"""

import collections
import collections.abc
import dataclasses
import datetime
import logging
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, TypeVar, Union

from volatility3.framework import interfaces
from volatility3.framework.interfaces import renderers

vollog = logging.getLogger(__name__)


class UnreadableValue(interfaces.renderers.BaseAbsentValue):
    """Class that represents values which are empty because the data cannot be
    read."""

    def __str__(self) -> str:
        """Fallback method for rendering basic types"""
        return "-"


class UnparsableValue(interfaces.renderers.BaseAbsentValue):
    """Class that represents values which are empty because the data cannot be
    interpreted correctly."""

    def __str__(self) -> str:
        """Fallback method for rendering basic types"""
        return "-"


class NotApplicableValue(interfaces.renderers.BaseAbsentValue):
    """Class that represents values which are empty because they don't make
    sense for this node."""

    def __str__(self) -> str:
        """Fallback method for rendering basic types"""
        return "N/A"


class NotAvailableValue(interfaces.renderers.BaseAbsentValue):
    """Class that represents values which cannot be provided now (but might in
    a future run)

    This might occur when information packed with volatility (such as
    symbol information) is not available, but a future version or a
    different run may later have that information available (ie, it
    could be applicable, but we can't get it and it's not because it's
    unreadable or unparsable). Unreadable and Unparsable should be used
    in preference, and only if neither fits should this be used.
    """

    def __str__(self) -> str:
        """Fallback method for rendering basic types"""
        return "N/A"


##########
### Basic Types


class Disassembly(interfaces.renderers.BasicType):
    """A class to indicate that the bytes provided should be disassembled
    (based on the architecture)"""

    possible_architectures = ["intel", "intel64", "arm", "arm64"]

    def __init__(
        self, data: bytes, offset: int = 0, architecture: str = "intel64"
    ) -> None:
        self.data = data
        self.architecture = None
        if architecture in self.possible_architectures:
            self.architecture = architecture
        if not isinstance(offset, int):
            raise TypeError("Offset must be an integer type")
        self.offset = offset

    def __str__(self) -> str:
        """Fallback method of rendering"""
        return str(self.data)


@dataclasses.dataclass
class LayerData(interfaces.renderers.BasicType):
    """Layer data

    This requires the context to be passed in, in case plugins want to use multiple contexts
    and to ensure the TreeGrid interface doesn't change, since this would break all existing plugins
    """

    context: "interfaces.context.ContextInterface"
    layer_name: str
    offset: int
    length: int
    no_surrounding: bool = False

    @staticmethod
    def from_object(
        object: "interfaces.objects.ObjectInterface",
        size: Optional[int] = None,
        no_surrounding: bool = True,
    ):
        return LayerData(
            context=object._context,
            layer_name=object.vol.layer_name,
            offset=object.vol.offset,
            length=size or object.vol.size,
            no_surrounding=no_surrounding,
        )

    def __str__(self) -> str:
        """Fallback method of rendering"""
        data = self.context.layers[self.layer_name].read(self.offset, self.length, True)
        return str(data)


class TreeNode(interfaces.renderers.TreeNode):
    """Class representing a particular node in a tree grid."""

    def __init__(
        self,
        path: str,
        treegrid: "TreeGrid",
        parent: Optional[interfaces.renderers.TreeNode],
        values: List[interfaces.renderers.BaseTypes],
    ) -> None:
        if not isinstance(treegrid, TreeGrid):
            raise TypeError("Treegrid must be an instance of TreeGrid")
        self._treegrid = treegrid
        self._parent = parent
        self._path = path
        self._validate_values(values)
        self._values = treegrid.RowStructure(*values)  # type: ignore

    def __repr__(self) -> str:
        return f"<TreeNode [{self.path}] - {self._values}>"

    def __getitem__(self, item: Union[int, slice]) -> Any:
        return self._treegrid.children(self).__getitem__(item)

    def __len__(self) -> int:
        return len(self._treegrid.children(self))

    def _validate_values(self, values: List[interfaces.renderers.BaseTypes]) -> None:
        """A function for raising exceptions if a given set of values is
        invalid according to the column properties."""
        if not (
            isinstance(values, collections.abc.Sequence)
            and len(values) == len(self._treegrid.columns)
        ):
            raise TypeError(
                "Values must be a list of objects made up of simple types and number the same as the columns"
            )
        for index, column in enumerate(self._treegrid.columns):
            val = values[index]
            if not isinstance(val, (column.type, interfaces.renderers.BaseAbsentValue)):
                raise TypeError(
                    f"Values item with index {index} is the wrong type for column {column.name} (got {type(val)} but expected {column.type})"
                )
            # TODO: Consider how to deal with timezone naive/aware datetimes (and alert plugin uses to be precise)
            # if isinstance(val, datetime.datetime):
            #     tznaive = val.tzinfo is None or val.tzinfo.utcoffset(val) is None

    def asdict(self) -> Dict[str, Any]:
        """Returns the contents of the node as a dictionary"""
        return self._values._asdict()

    @property
    def values(self) -> List[interfaces.renderers.BaseTypes]:
        """Returns the list of values from the particular node, based on column
        index."""
        return list(self._values)

    @property
    def path(self) -> str:
        """Returns a path identifying string.

        This should be seen as opaque by external classes, Parsing of
        path locations based on this string are not guaranteed to remain
        stable.
        """
        return self._path

    @property
    def parent(self) -> Optional[interfaces.renderers.TreeNode]:
        """Returns the parent node of this node or None."""
        return self._parent

    @property
    def path_depth(self) -> int:
        """Return the path depth of the current node."""
        return len(self.path.split(TreeGrid.path_sep))

    def path_changed(self, path: str, added: bool = False) -> None:
        """Updates the path based on the addition or removal of a node higher
        up in the tree.

        This should only be called by the containing TreeGrid and
        expects to only be called for affected nodes.
        """
        components = self._path.split(TreeGrid.path_sep)
        changed = path.split(TreeGrid.path_sep)
        changed_index = len(changed) - 1
        if int(components[changed_index]) >= int(changed[-1]):
            components[changed_index] = str(
                int(components[changed_index]) + (1 if added else -1)
            )
        self._path = TreeGrid.path_sep.join(components)


def RowStructureConstructor(names: List[str]):
    return collections.namedtuple(
        "RowStructure", [TreeGrid.sanitize_name(name) for name in names]
    )


class TreeGrid(interfaces.renderers.TreeGrid):
    """Class providing the interface for a TreeGrid (which contains TreeNodes)

    The structure of a TreeGrid is designed to maintain the structure of the tree in a single object.
    For this reason each TreeNode does not hold its children, they are managed by the top level object.
    This leaves the Nodes as simple data carries and prevents them being used to manipulate the tree as a whole.
    This is a data structure, and is not expected to be modified much once created.

    Carrying the children under the parent makes recursion easier, but then every node is its own little tree
    and must have all the supporting tree functions.  It also allows for a node to be present in several different trees,
    and to create cycles.
    """

    path_sep = "|"

    def __init__(
        self,
        columns: List[Tuple[str, interfaces.renderers.BaseTypes]],
        generator: Optional[Iterable[Tuple[int, Tuple]]],
    ) -> None:
        """Constructs a TreeGrid object using a specific set of columns.

        The TreeGrid itself is a root element, that can have children but no values.
        The TreeGrid does *not* contain any information about formatting,
        these are up to the renderers and plugins.

        Args:
            columns: A list of column tuples made up of (name, type).
            generator: An iterable containing row for a tree grid, each row contains a indent level followed by the values for each column in order.
        """
        self._populated = False
        self._row_count = 0
        self._children: List[interfaces.renderers.TreeNode] = []
        converted_columns: List[interfaces.renderers.Column] = []
        if len(columns) < 1:
            raise ValueError("Columns must be a list containing at least one column")
        for name, column_type in columns:
            is_simple_type = issubclass(column_type, self.base_types)
            if not is_simple_type:
                raise TypeError(
                    f"Column {name}'s type is not a simple type: {column_type.__class__.__name__}"
                )
            converted_columns.append(interfaces.renderers.Column(name, column_type))
        self.RowStructure = RowStructureConstructor(
            [column.name for column in converted_columns]
        )
        self._columns = converted_columns
        if generator is None:
            generator = []
        generator = iter(generator)

        self._generator = generator

    @staticmethod
    def sanitize_name(text: str) -> str:
        output = ""
        for letter in text.lower():
            if letter != " ":
                output += (
                    letter
                    if letter in "abcdefghiljklmnopqrstuvwxyz_0123456789"
                    else "_"
                )
        return output

    def populate(
        self,
        function: Optional[interfaces.renderers.VisitorSignature] = None,
        initial_accumulator: Any = None,
        fail_on_errors: bool = True,
    ) -> Optional[Exception]:
        """Populates the tree by consuming the TreeGrid's construction
        generator Func is called on every node, so can be used to create output
        on demand.

        This is equivalent to a one-time visit.

        Args:
            function: The visitor to be called on each row of the treegrid
            initial_accumulator: The initial value for an accumulator passed to the visitor to allow it to maintain state
            fail_on_errors: A boolean defining whether exceptions should be caught or bubble up
        """
        accumulator = initial_accumulator
        if function is None:

            def function(_x: interfaces.renderers.TreeNode, _y: Any) -> Any:
                return None

        if not self.populated:
            try:
                prev_nodes: List[interfaces.renderers.TreeNode] = []
                for level, item in self._generator:
                    parent_index = min(len(prev_nodes), level)
                    parent = prev_nodes[parent_index - 1] if parent_index > 0 else None
                    treenode = self._append(parent, item)
                    prev_nodes = prev_nodes[0:parent_index] + [treenode]
                    if function is not None:
                        accumulator = function(treenode, accumulator)
                    self._row_count += 1
            except Exception as excp:
                if fail_on_errors:
                    raise
                vollog.debug(f"Exception during population: {excp}")
                self._populated = True
                return excp
        self._populated = True
        return None

    @property
    def populated(self) -> bool:
        """Indicates that population has completed and the tree may now be
        manipulated separately."""
        return self._populated

    @property
    def columns(self) -> List[interfaces.renderers.Column]:
        """Returns the available columns and their ordering and types."""
        return self._columns

    @property
    def row_count(self) -> int:
        """Returns the number of rows populated."""
        return self._row_count

    def children(
        self, node: Optional[interfaces.renderers.TreeNode]
    ) -> List[interfaces.renderers.TreeNode]:
        """Returns the subnodes of a particular node in order."""
        return [node for node, _ in self._find_children(node)]

    def _find_children(self, node: Optional[interfaces.renderers.TreeNode]) -> Any:
        """Returns the children list associated with a particular node.

        Returns None if the node does not exist
        """
        children = self._children
        try:
            if node is not None:
                for path_component in node.path.split(self.path_sep):
                    _, children = children[int(path_component)]
        except IndexError:
            return []
        return children

    def values(self, node):
        """Returns the values for a particular node.

        The values returned are mutable,
        """
        if node is None:
            raise TypeError("Node must be a valid node within the TreeGrid")
        return node.values

    def _append(
        self, parent: Optional[interfaces.renderers.TreeNode], values: Any
    ) -> TreeNode:
        """Adds a new node at the top level if parent is None, or under the
        parent node otherwise, after all other children."""
        return self._insert(parent, None, values)

    def _insert(
        self,
        parent: Optional[interfaces.renderers.TreeNode],
        position: Optional[int],
        values: Any,
    ) -> TreeNode:
        """Inserts an element into the tree at a specific position."""
        parent_path = ""
        children = self._find_children(parent)
        if parent is not None:
            parent_path = parent.path + self.path_sep
        if position is None:
            newpath = parent_path + str(len(children))
        else:
            newpath = parent_path + str(position)
            for node, _ in children[position:]:
                self.visit(
                    node, lambda child, _: child.path_changed(newpath, True), None
                )

        tree_item = TreeNode(newpath, self, parent, values)
        if position is None:
            children.append((tree_item, []))
        else:
            children.insert(position, (tree_item, []))
        return tree_item

    def is_ancestor(self, node, descendant):
        """Returns true if descendent is a child, grandchild, etc of node."""
        return descendant.path.startswith(node.path)

    def max_depth(self):
        """Returns the maximum depth of the tree."""
        return self.visit(None, lambda n, a: max(a, self.path_depth(n)), 0)

    _T = TypeVar("_T")

    def visit(
        self,
        node: Optional[interfaces.renderers.TreeNode],
        function: Callable[[interfaces.renderers.TreeNode, _T], _T],
        initial_accumulator: _T,
        sort_key: Optional[interfaces.renderers.ColumnSortKey] = None,
    ):
        """Visits all the nodes in a tree, calling function on each one.

        function should have the signature function(node, accumulator) and return new_accumulator
        If accumulators are not needed, the function must still accept a second parameter.

        The order of that the nodes are visited is always depth first, however, the order children are traversed can
        be set based on a sort_key function which should accept a node's values and return something that can be
        sorted to receive the desired order (similar to the sort/sorted key).

        We use the private _find_children function so that we don't have to re-traverse the tree
        for every node we descend further down
        """
        if not self.populated:
            self.populate()

        # Find_nodes is path dependent, whereas _visit is not
        # So in case the function modifies the node's path, find the nodes first
        children = self._find_children(node)
        accumulator = initial_accumulator
        # We split visit into two, so that we don't have to keep calling find_children to traverse the tree
        if node is not None:
            accumulator = function(node, initial_accumulator)
        if children is not None:
            if sort_key is not None:
                sort_key_not_none = sort_key  # Only necessary because of mypy
                children = sorted(
                    children, key=lambda x: sort_key_not_none(x[0].values)
                )
                if not sort_key.ascending:
                    children = reversed(children)
            accumulator = self._visit(children, function, accumulator, sort_key)
        return accumulator

    def _visit(
        self,
        list_of_children: List[interfaces.renderers.TreeNode],
        function: Callable,
        accumulator: _T,
        sort_key: Optional[interfaces.renderers.ColumnSortKey] = None,
    ) -> _T:
        """Visits all the nodes in a tree, calling function on each one."""
        if list_of_children is not None:
            for n, children in list_of_children:
                accumulator = function(n, accumulator)
                if sort_key is not None:
                    sort_key_not_none = sort_key  # Only necessary because of mypy
                    children = sorted(
                        children, key=lambda x: sort_key_not_none(x[0].values)
                    )
                    if not sort_key.ascending:
                        children = reversed(children)
                accumulator = self._visit(children, function, accumulator, sort_key)
        return accumulator


class ColumnSortKey(interfaces.renderers.ColumnSortKey):
    def __init__(
        self, treegrid: TreeGrid, column_name: str, ascending: bool = True
    ) -> None:
        _index = None
        self._type = None
        self.ascending = ascending
        for i, column in enumerate(treegrid.columns):
            if column.name.lower() == column_name.lower():
                _index = i
                self._type = column.type
        if _index is None:
            raise ValueError(f"Column not found in TreeGrid columns: {column_name}")
        self._index = _index

    def __call__(self, values: List[Any]) -> Any:
        """The key function passed as the sort key."""
        value = values[self._index]
        if isinstance(value, interfaces.renderers.BaseAbsentValue):
            if self._type == datetime.datetime:
                value = datetime.datetime.min
            elif self._type in [int, float]:
                value = -1
            elif self._type is bool:
                value = False
            elif self._type in [str, renderers.Disassembly]:
                value = "-"
            elif self._type is bytes:
                value = b""
        return value



# This file is Copyright 2019 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#
import contextlib
import enum
import logging
import struct
from typing import Iterator, Optional, Union, cast

from volatility3.framework import constants, exceptions, interfaces, objects
from volatility3.framework.layers import registry

vollog = logging.getLogger(__name__)

BIG_DATA_MAXLEN = 0x3FD8


class RegValueTypes(enum.Enum):
    REG_NONE = 0
    REG_SZ = 1
    REG_EXPAND_SZ = 2
    REG_BINARY = 3
    REG_DWORD = 4
    REG_DWORD_BIG_ENDIAN = 5
    REG_LINK = 6
    REG_MULTI_SZ = 7
    REG_RESOURCE_LIST = 8
    REG_FULL_RESOURCE_DESCRIPTOR = 9
    REG_RESOURCE_REQUIREMENTS_LIST = 10
    REG_QWORD = 11
    REG_UNKNOWN = 99999

    @classmethod
    def _missing_(cls, value):
        return cls(RegValueTypes.REG_UNKNOWN)


INTEGER_TYPES = [
    RegValueTypes.REG_DWORD,
    RegValueTypes.REG_QWORD,
    RegValueTypes.REG_DWORD_BIG_ENDIAN,
    RegValueTypes.REG_DWORD_BIG_ENDIAN,
]

STRING_TYPES = [
    RegValueTypes.REG_SZ,
    RegValueTypes.REG_MULTI_SZ,
    RegValueTypes.REG_EXPAND_SZ,
    RegValueTypes.REG_LINK,
]

BINARY_TYPES = [
    RegValueTypes.REG_RESOURCE_LIST,
    RegValueTypes.REG_BINARY,
    RegValueTypes.REG_FULL_RESOURCE_DESCRIPTOR,
    RegValueTypes.REG_RESOURCE_REQUIREMENTS_LIST,
    RegValueTypes.REG_NONE,
]


class RegKeyFlags(enum.IntEnum):
    KEY_IS_VOLATILE = 0x01
    KEY_HIVE_EXIT = 0x02
    KEY_HIVE_ENTRY = 0x04
    KEY_NO_DELETE = 0x08
    KEY_SYM_LINK = 0x10
    KEY_COMP_NAME = 0x20
    KEY_PREFEF_HANDLE = 0x40
    KEY_VIRT_MIRRORED = 0x80
    KEY_VIRT_TARGET = 0x100
    KEY_VIRTUAL_STORE = 0x200


class HMAP_ENTRY(objects.StructType):
    def get_block_offset(self) -> int:
        try:
            return (
                self.PermanentBinAddress ^ (self.PermanentBinAddress & 0xF)
            ) + self.BlockOffset
        except AttributeError:
            return self.BlockAddress


class CMHIVE(objects.StructType):
    def is_valid(self) -> bool:
        """Determine if the object is valid."""
        try:
            return self.Hive.Signature == 0xBEE0BEE0
        except exceptions.InvalidAddressException:
            return False

    def get_name(self) -> Optional[interfaces.objects.ObjectInterface]:
        """Determine a name for the hive.

        Note that some attributes are unpredictably blank across
        different OS versions while others are populated, so we check
        all possibilities and take the first one that's not empty
        """

        for attr in ["FileFullPath", "FileUserName", "HiveRootPath"]:
            with contextlib.suppress(
                AttributeError,
                exceptions.InvalidAddressException,
                registry.RegistryException,
            ):
                name = getattr(self, attr)
                if name.Length > 0:
                    return name.get_string()

        return None

    name = property(get_name)


class CM_KEY_BODY(objects.StructType):
    """This represents an open handle to a registry key and is not tied to the
    registry hive file format on disk."""

    def _skip_key_hive_entry_path(self, kcb_flags):
        """Win10 14393 introduced an extra path element that it skips over by
        checking for Flags that contain KEY_HIVE_ENTRY."""

        # _CM_KEY_BODY.Trans introduced in Win10 14393
        if (
            hasattr(self, "Trans")
            and RegKeyFlags.KEY_HIVE_ENTRY & kcb_flags == RegKeyFlags.KEY_HIVE_ENTRY
        ):
            return True

        return False

    def get_full_key_name(self) -> str:
        output = []
        seen = set()

        kcb = self.KeyControlBlock
        while kcb.ParentKcb:
            if kcb.ParentKcb.vol.offset in seen:
                return None
            seen.add(kcb.ParentKcb.vol.offset)

            if len(output) > 128:
                return None

            if kcb.NameBlock.Name is None:
                break

            if self._skip_key_hive_entry_path(kcb.Flags):
                kcb = kcb.ParentKcb
                if not kcb:
                    break

            output.append(
                kcb.NameBlock.Name.cast(
                    "string",
                    encoding="utf8",
                    max_length=kcb.NameBlock.NameLength,
                    errors="replace",
                )
            )
            kcb = kcb.ParentKcb
        return "\\".join(reversed(output))


class CM_KEY_NODE(objects.StructType):
    """Extension to allow traversal of registry keys."""

    def get_volatile(self) -> bool:
        """
        Returns a bool indicating whether or not the key is volatile.

        Raises TypeError if the key was not instantiated on a RegistryHive layer
        """
        if not isinstance(
            self._context.layers[self.vol.layer_name], registry.RegistryHive
        ):
            raise TypeError("CM_KEY_NODE was not instantiated on a RegistryHive layer")
        return bool(self.vol.offset & 0x80000000)

    def get_subkeys(self) -> Iterator["CM_KEY_NODE"]:
        """Returns a list of the key nodes.

        Raises TypeError if the key was not instantiated on a RegistryHive layer
        """
        hive = self._context.layers[self.vol.layer_name]
        if not isinstance(hive, registry.RegistryHive):
            raise TypeError("CM_KEY_NODE was not instantiated on a RegistryHive layer")
        for index in range(2):
            # Use get_cell because it should *always* be a KeyIndex
            subkey_node = hive.get_cell(self.SubKeyLists[index]).u.KeyIndex
            yield from self._get_subkeys_recursive(hive, subkey_node)

    def _get_subkeys_recursive(
        self, hive: "registry.RegistryHive", node: interfaces.objects.ObjectInterface
    ) -> Iterator["CM_KEY_NODE"]:
        """Recursively descend a node returning subkeys."""
        # The keylist appears to include 4 bytes of key name after each value
        # We can either double the list and only use the even items, or
        # We could change the array type to a struct with both parts
        try:
            signature = node.cast("string", max_length=2, encoding="latin-1")
        except (
            exceptions.InvalidAddressException,
            registry.RegistryException,
        ):
            return None

        listjump = None
        if signature == "ri":
            listjump = 1
        elif signature == "lh" or signature == "lf":
            listjump = 2
        elif node.vol.type_name.endswith(constants.BANG + "_CM_KEY_NODE"):
            yield cast("CM_KEY_NODE", node)
        else:
            vollog.debug(
                f"Unexpected node type encountered when traversing subkeys: {node.vol.type_name}, signature: {signature}"
            )

        if listjump:
            node.List.count = node.Count * listjump
            for subnode_offset in node.List[::listjump]:
                if (subnode_offset & 0x7FFFFFFF) > hive.maximum_address:
                    vollog.log(
                        constants.LOGLEVEL_VVV,
                        f"Node found with address outside the valid Hive size: {hex(subnode_offset)}",
                    )
                else:
                    try:
                        subnode = hive.get_node(subnode_offset)
                    except (
                        exceptions.InvalidAddressException,
                        registry.RegistryException,
                    ):
                        vollog.log(
                            constants.LOGLEVEL_VVV,
                            f"Failed to get node at {hex(subnode_offset)}, skipping",
                        )
                        continue
                    yield from self._get_subkeys_recursive(hive, subnode)

    def get_values(self) -> Iterator["CM_KEY_VALUE"]:
        """Returns a list of the Value nodes for a key.

        Raises TypeError if the key was not instantiated on a RegistryHive layer
        """
        hive = self._context.layers[self.vol.layer_name]
        if not isinstance(hive, registry.RegistryHive):
            raise TypeError("CM_KEY_NODE was not instantiated on a RegistryHive layer")

        try:
            child_list = hive.get_cell(self.ValueList.List).u.KeyList
            child_list.count = self.ValueList.Count

            for v in child_list:
                if v != 0:
                    try:
                        node = hive.get_node(v)
                    except (registry.RegistryException,) as excp:
                        vollog.debug(f"Invalid address {excp}")
                        continue
                    if isinstance(node, CM_KEY_VALUE):
                        yield node

        except (
            exceptions.InvalidAddressException,
            registry.RegistryException,
        ) as excp:
            vollog.debug(f"Invalid address in get_values iteration: {excp}")
            return None

    def get_name(self) -> interfaces.objects.ObjectInterface:
        """Gets the name for the current key node"""
        namelength = self.NameLength
        self.Name.count = namelength
        return self.Name.cast("string", max_length=namelength, encoding="latin-1")

    def get_key_path(self) -> str:
        """
        Returns the full path to this registry key.

        Raises TypeError if the key was not instantiated on a RegistryHive layer
        """
        reg = self._context.layers[self.vol.layer_name]
        if not isinstance(reg, registry.RegistryHive):
            raise TypeError("Key was not instantiated on a RegistryHive layer")
        # Using the offset adds a significant delay (since it cannot be cached easily)
        # if self.vol.offset == reg.get_node(reg.root_cell_offset).vol.offset:
        if self.vol.offset == reg.root_cell_offset + 4:
            # return the last part of the hive name for the root entry
            return reg.get_name().split("\\")[-1]
        return reg.get_node(self.Parent).get_key_path() + "\\" + self.get_name()


class CM_KEY_VALUE(objects.StructType):
    """Extensions to extract data from CM_KEY_VALUE nodes."""

    def get_name(self) -> interfaces.objects.ObjectInterface:
        """Gets the name for the current key value"""
        namelength = self.NameLength
        self.Name.count = namelength
        return self.Name.cast("string", max_length=namelength, encoding="latin-1")

    def get_type(self) -> RegValueTypes:
        """Get the type of the registry value"""
        return RegValueTypes(self.Type)

    def decode_data(self) -> Union[int, bytes]:
        """
        Properly decodes the data associated with the value node.

        If an InvalidAddressException occurs when reading data from the
        underlying RegistryHive layer, the data will be padded with null bytes
        of the same length.

        Raises ValueError if the data cannot be read
        Raises TypeError if the class was not instantiated on a RegistryHive layer
        """
        # Determine if the data is stored inline
        datalen = self.DataLength
        data = b""
        # Check if the data is stored inline
        layer = self._context.layers[self.vol.layer_name]
        if not isinstance(layer, registry.RegistryHive):
            raise TypeError("Key value was not instantiated on a RegistryHive layer")

        # If the high-bit is set
        if datalen & 0x80000000:
            # Remove the high bit
            datalen = datalen & 0x7FFFFFFF
            if 0 > datalen or datalen > 4:
                raise ValueError(
                    f"Unable to read inline registry value with excessive length: {datalen}"
                )
            else:
                data = layer.read(self.Data.vol.offset, datalen)
        elif layer.hive.Version == 5 and datalen > 0x4000:
            # We're bigdata
            big_data = layer.get_node(self.Data).cast("_CM_BIG_DATA")
            # Oddly, we get a list of addresses, at which are addresses, which then point to data blocks
            for i in range(big_data.Count):
                # The value 4 should actually be unsigned-int.size, but since it's a file format that shouldn't change
                # the direct value 4 can be used instead
                block_offset = layer.get_cell(big_data.List + (i * 4)).cast(
                    "unsigned int"
                )
                if (
                    isinstance(block_offset, int)
                    and block_offset < layer.maximum_address
                ):
                    amount = min(BIG_DATA_MAXLEN, datalen)
                    try:
                        data += layer.read(
                            offset=layer.get_cell(block_offset).vol.offset,
                            length=amount,
                        )
                    except (
                        exceptions.InvalidAddressException,
                        registry.RegistryException,
                    ):
                        vollog.debug(
                            f"Failed to read {amount:x} bytes of data, padding with {amount:x}"
                        )
                    datalen -= amount
        else:
            # Suspect Data actually points to a Cell,
            # but the length at the start could be negative so just adding 4 to jump past it
            try:
                data = layer.read(self.Data + 4, datalen)
            except (exceptions.InvalidAddressException, registry.RegistryException):
                vollog.debug(
                    f"Failed to read {datalen:x} bytes of data, returning {datalen:x} null bytes"
                )
                data = b"\x00" * datalen

        if self.get_type() == RegValueTypes.REG_DWORD:
            if len(data) != struct.calcsize("<L"):
                raise ValueError(
                    f"Size of data does not match the type of registry value {self.get_name()}"
                )
            (res,) = struct.unpack("<L", data)
            return res
        if self.get_type() == RegValueTypes.REG_DWORD_BIG_ENDIAN:
            if len(data) != struct.calcsize(">L"):
                raise ValueError(
                    f"Size of data does not match the type of registry value {self.get_name()}"
                )
            (res,) = struct.unpack(">L", data)
            return res
        if self.get_type() == RegValueTypes.REG_QWORD:
            if len(data) != struct.calcsize("<Q"):
                raise ValueError(
                    f"Size of data does not match the type of registry value {self.get_name()}"
                )
            (res,) = struct.unpack("<Q", data)
            return res
        if self.get_type() in [
            RegValueTypes.REG_SZ,
            RegValueTypes.REG_EXPAND_SZ,
            RegValueTypes.REG_LINK,
            RegValueTypes.REG_MULTI_SZ,
            RegValueTypes.REG_BINARY,
            RegValueTypes.REG_FULL_RESOURCE_DESCRIPTOR,
            RegValueTypes.REG_RESOURCE_LIST,
            RegValueTypes.REG_RESOURCE_REQUIREMENTS_LIST,
        ]:
            return data
        if self.get_type() == RegValueTypes.REG_NONE:
            return b""

        # Fall back if it's something weird
        vollog.debug(f"Unknown registry value type encountered: {self.Type}")
        return data



# This file is Copyright 2019 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#
"""The official list of format hints that text renderers and plugins can rely
upon existing within the framework.

These hints allow a plugin to indicate how they would like data from a particular column to be represented.

Text renderers should attempt to honour all hints provided in this module where possible
"""

from typing import Type, Union

from volatility3.framework import interfaces


class Bin(int):
    """A class to indicate that the integer value should be represented as a
    binary value."""


class Hex(int):
    """A class to indicate that the integer value should be represented as a
    hexadecimal value."""


class HexBytes(bytes):
    """A class to indicate that the bytes should be display in an extended
    format showing hexadecimal and ascii printable display."""


class MultiTypeData(bytes):
    """The contents are supposed to be a string, but may contain binary data."""

    def __new__(
        cls: Type["MultiTypeData"],
        original: Union[int, bytes],
        encoding: str = "utf-16-le",
        split_nulls: bool = False,
        show_hex: bool = False,
    ) -> "MultiTypeData":
        if isinstance(original, int):
            data = str(original).encode(encoding)
        else:
            data = original
        return super().__new__(cls, data)

    def __init__(
        self,
        original: bytes,
        encoding: str = "utf-16-le",
        split_nulls: bool = False,
        show_hex: bool = False,
    ) -> None:
        self.converted_int: bool = False
        if isinstance(original, int):
            self.converted_int = True
        self.encoding = encoding
        self.split_nulls = split_nulls
        self.show_hex = show_hex
        bytes.__init__(original)

    def __eq__(self, other):
        return (
            isinstance(other, self.__class__)
            and super() == super(self.__class__, other)
            and self.converted_int == other.converted_int
            and self.encoding == other.encoding
            and self.split_nulls == other.split_nulls
            and self.show_hex == other.show_hex
        )


def BinOrAbsent(x):
    return Bin(x) if not isinstance(x, interfaces.renderers.BaseAbsentValue) else x


def HexOrAbsent(x):
    return Hex(x) if not isinstance(x, interfaces.renderers.BaseAbsentValue) else x


def HexBytesOrAbsent(x):
    return HexBytes(x) if not isinstance(x, interfaces.renderers.BaseAbsentValue) else x


def MultiTypeDataOrAbsent(x):
    return (
        MultiTypeData(x)
        if not isinstance(x, interfaces.renderers.BaseAbsentValue)
        else x
    )



# This file is Copyright 2024 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#
import contextlib
import datetime
import logging
import re
from typing import List

from volatility3.framework import exceptions, interfaces, renderers
from volatility3.framework.configuration import requirements
from volatility3.framework.renderers import conversion, format_hints
from volatility3.framework.symbols import intermed
from volatility3.framework.symbols.windows.extensions import pe
from volatility3.plugins import timeliner
from volatility3.plugins.windows import info, pedump, pslist, psscan

vollog = logging.getLogger(__name__)


class DllList(interfaces.plugins.PluginInterface, timeliner.TimeLinerInterface):
    """Lists the loaded DLLs in a particular windows memory image."""

    _required_framework_version = (2, 0, 0)
    _version = (3, 0, 1)

    @classmethod
    def get_requirements(cls) -> List[interfaces.configuration.RequirementInterface]:
        # Since we're calling the plugin, make sure we have the plugin's requirements
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Windows kernel",
                architectures=["Intel32", "Intel64"],
            ),
            requirements.VersionRequirement(
                name="pslist", component=pslist.PsList, version=(3, 0, 0)
            ),
            requirements.VersionRequirement(
                name="timeliner",
                component=timeliner.TimeLinerInterface,
                version=(1, 0, 0),
            ),
            requirements.VersionRequirement(
                name="psscan", component=psscan.PsScan, version=(2, 0, 0)
            ),
            requirements.VersionRequirement(
                name="pedump", component=pedump.PEDump, version=(2, 0, 0)
            ),
            requirements.VersionRequirement(
                name="info", component=info.Info, version=(2, 0, 0)
            ),
            requirements.ListRequirement(
                name="pid",
                element_type=int,
                description="Process IDs to include (all other processes are excluded)",
                optional=True,
            ),
            requirements.IntRequirement(
                name="offset",
                description="Process offset in the physical address space",
                optional=True,
            ),
            requirements.IntRequirement(
                name="base",
                description="Specify a base virtual address in process memory",
                optional=True,
            ),
            requirements.StringRequirement(
                name="name",
                description="Specify a regular expression to match dll name(s)",
                optional=True,
            ),
            requirements.BooleanRequirement(
                name="ignore-case",
                description="Specify case insensitivity for the regular expression name matching",
                default=False,
                optional=True,
            ),
            requirements.BooleanRequirement(
                name="dump",
                description="Extract listed DLLs",
                default=False,
                optional=True,
            ),
        ]

    def _generator(self, procs):
        pe_table_name = intermed.IntermediateSymbolTable.create(
            self.context, self.config_path, "windows", "pe", class_types=pe.class_types
        )

        kuser = info.Info.get_kuser_structure(self.context, self.config["kernel"])

        nt_major_version = int(kuser.NtMajorVersion)
        nt_minor_version = int(kuser.NtMinorVersion)

        # LoadTime only applies to versions higher or equal to Window 7 (6.1 and higher)
        dll_load_time_field = (nt_major_version > 6) or (
            nt_major_version == 6 and nt_minor_version >= 1
        )

        for proc in procs:
            proc_id = proc.UniqueProcessId
            proc_layer_name = proc.add_process_layer()

            for entry in proc.load_order_modules():
                BaseDllName = FullDllName = renderers.UnreadableValue()
                with contextlib.suppress(exceptions.InvalidAddressException):
                    BaseDllName = entry.BaseDllName.get_string()
                    # We assume that if BaseDllName points to invalid buffer, so will FullDllName
                    FullDllName = entry.FullDllName.get_string()

                # Check if a name regex was passed and apply it to only show matches
                if self.config["name"]:
                    try:
                        flags = re.I if self.config["ignore-case"] else 0
                        mod_re = re.compile(self.config["name"], flags)
                    except re.error:
                        vollog.debug(
                            "Error parsing regular expression: %s", self.config["name"]
                        )
                        return None

                    # If Base or Full Dll Name are invalid, move on
                    if isinstance(BaseDllName, renderers.UnreadableValue) or isinstance(
                        FullDllName, renderers.UnreadableValue
                    ):
                        continue

                    # If regex does not match, move on
                    if not mod_re.search(BaseDllName) and not mod_re.search(
                        FullDllName
                    ):
                        continue

                if self.config["base"] and self.config["base"] != entry.DllBase:
                    continue

                if dll_load_time_field:
                    # Versions prior to 6.1 won't have the LoadTime attribute
                    # and 32-bit version shouldn't have the Quadpart according to MSDN
                    try:
                        DllLoadTime = conversion.wintime_to_datetime(
                            entry.LoadTime.QuadPart
                        )
                    except exceptions.InvalidAddressException:
                        DllLoadTime = renderers.UnreadableValue()
                else:
                    DllLoadTime = renderers.NotApplicableValue()

                file_output = "Disabled"
                if self.config["dump"]:
                    file_output = pedump.PEDump.dump_ldr_entry(
                        self.context,
                        pe_table_name,
                        entry,
                        self.open,
                        proc_layer_name,
                        prefix=f"pid.{proc_id}.",
                    )

                    if not file_output:
                        file_output = "Error outputting file"

                try:
                    dllbase = format_hints.Hex(entry.DllBase)
                except exceptions.InvalidAddressException:
                    dllbase = renderers.NotAvailableValue()

                try:
                    size_of_image = format_hints.Hex(entry.SizeOfImage)
                except exceptions.InvalidAddressException:
                    size_of_image = renderers.NotAvailableValue()

                LoadCount = entry.get_load_count()
                if LoadCount is None:
                    LoadCount = renderers.NotAvailableValue()

                yield (
                    0,
                    (
                        proc.UniqueProcessId,
                        proc.ImageFileName.cast(
                            "string",
                            max_length=proc.ImageFileName.vol.count,
                            errors="replace",
                        ),
                        dllbase,
                        size_of_image,
                        BaseDllName,
                        FullDllName,
                        LoadCount,
                        DllLoadTime,
                        file_output,
                    ),
                )

    def generate_timeline(self):
        for row in self._generator(
            pslist.PsList.list_processes(
                context=self.context, kernel_module_name=self.config["kernel"]
            )
        ):
            _depth, row_data = row
            if not isinstance(row_data[6], datetime.datetime):
                continue
            description = f"DLL Load: Process {row_data[0]} {row_data[1]} Loaded {row_data[4]} ({row_data[5]}) Size {row_data[3]} Offset {row_data[2]}"
            yield (description, timeliner.TimeLinerType.CREATED, row_data[6])

    def run(self):
        filter_func = pslist.PsList.create_pid_filter(self.config.get("pid", None))
        kernel = self.context.modules[self.config["kernel"]]

        if self.config["offset"]:
            procs = psscan.PsScan.scan_processes(
                self.context,
                self.config["kernel"],
                filter_func=psscan.PsScan.create_offset_filter(
                    self.context,
                    kernel.layer_name,
                    self.config["offset"],
                ),
            )
        else:
            procs = pslist.PsList.list_processes(
                context=self.context,
                kernel_module_name=self.config["kernel"],
                filter_func=filter_func,
            )

        return renderers.TreeGrid(
            [
                ("PID", int),
                ("Process", str),
                ("Base", format_hints.Hex),
                ("Size", format_hints.Hex),
                ("Name", str),
                ("Path", str),
                ("LoadCount", int),
                ("LoadTime", datetime.datetime),
                ("File output", str),
            ],
            self._generator(procs=procs),
        )



# This file is Copyright 2020 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#
import logging
import os
from datetime import datetime
from itertools import count
from typing import Iterator, List, Optional, Tuple

from volatility3.framework import constants, exceptions, interfaces, renderers, symbols
from volatility3.framework.configuration import requirements
from volatility3.framework.objects.utility import array_to_string
from volatility3.framework.renderers import format_hints
from volatility3.framework.symbols import intermed
from volatility3.framework.symbols.windows import versions
from volatility3.framework.symbols.windows.extensions import pe, shimcache
from volatility3.plugins import timeliner
from volatility3.plugins.windows import modules, pslist, vadinfo

vollog = logging.getLogger(__name__)


class ShimcacheMem(interfaces.plugins.PluginInterface, timeliner.TimeLinerInterface):
    """Reads Shimcache entries from the ahcache.sys AVL tree"""

    _required_framework_version = (2, 0, 0)
    _version = (1, 0, 1)

    # These checks must be completed from newest -> oldest OS version.
    _win_version_file_map: List[Tuple[versions.OsDistinguisher, bool, str]] = [
        (versions.is_win10, True, "shimcache-win10-x64"),
        (versions.is_win10, False, "shimcache-win10-x86"),
        (versions.is_windows_8_or_later, True, "shimcache-win8-x64"),
        (versions.is_windows_8_or_later, False, "shimcache-win8-x86"),
        (versions.is_windows_7, True, "shimcache-win7-x64"),
        (versions.is_windows_7, False, "shimcache-win7-x86"),
        (versions.is_vista_or_later, True, "shimcache-vista-x64"),
        (versions.is_vista_or_later, False, "shimcache-vista-x86"),
        (versions.is_2003, False, "shimcache-2003-x86"),
        (versions.is_2003, True, "shimcache-2003-x64"),
        (versions.is_windows_xp_sp3, False, "shimcache-xp-sp3-x86"),
        (versions.is_windows_xp_sp2, False, "shimcache-xp-sp2-x86"),
        (versions.is_xp_or_2003, True, "shimcache-xp-2003-x64"),
        (versions.is_xp_or_2003, False, "shimcache-xp-2003-x86"),
    ]

    NT_KRNL_MODS = ["ntoskrnl.exe", "ntkrnlpa.exe", "ntkrnlmp.exe", "ntkrpamp.exe"]

    def generate_timeline(
        self,
    ) -> Iterator[Tuple[str, timeliner.TimeLinerType, datetime]]:
        for _, (_, last_modified, last_update, _, _, file_path) in self._generator():
            if isinstance(last_update, datetime):
                yield (
                    f"Shimcache: File {file_path} executed",
                    timeliner.TimeLinerType.ACCESSED,
                    last_update,
                )
            if isinstance(last_modified, datetime):
                yield (
                    f"Shimcache: File {file_path} modified",
                    timeliner.TimeLinerType.MODIFIED,
                    last_modified,
                )

    @classmethod
    def get_requirements(cls) -> List[interfaces.configuration.RequirementInterface]:
        # Since we're calling the plugin, make sure we have the plugin's requirements
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Windows kernel",
                architectures=["Intel32", "Intel64"],
            ),
            requirements.VersionRequirement(
                name="pslist", component=pslist.PsList, version=(3, 0, 0)
            ),
            requirements.VersionRequirement(
                name="timeliner",
                component=timeliner.TimeLinerInterface,
                version=(1, 0, 0),
            ),
            requirements.VersionRequirement(
                name="vadinfo", component=vadinfo.VadInfo, version=(2, 0, 0)
            ),
            requirements.VersionRequirement(
                name="modules", component=modules.Modules, version=(3, 0, 0)
            ),
        ]

    @classmethod
    def create_shimcache_table(
        cls,
        context: interfaces.context.ContextInterface,
        symbol_table_name: str,
        config_path: str,
    ) -> str:
        """Creates a shimcache symbol table

        Args:
            context: The context to retrieve required elements (layers, symbol tables) from
            symbol_table: The name of an existing symbol table containing the kernel symbols
            config_path: The configuration path within the context of the symbol table to create

        Returns:
            The name of the constructed shimcache table
        """
        native_types = context.symbol_space[symbol_table_name].natives
        is_64bit = symbols.symbol_table_is_64bit(
            context=context, symbol_table_name=symbol_table_name
        )
        table_mapping = {"nt_symbols": symbol_table_name}

        try:
            symbol_filename = next(
                filename
                for version_check, for_64bit, filename in ShimcacheMem._win_version_file_map
                if is_64bit == for_64bit
                and version_check(context=context, symbol_table=symbol_table_name)
            )
        except StopIteration:
            raise NotImplementedError("This version of Windows is not supported!")

        vollog.debug(f"Using shimcache table {symbol_filename}")

        return intermed.IntermediateSymbolTable.create(
            context,
            config_path,
            os.path.join("windows", "shimcache"),
            symbol_filename,
            class_types=shimcache.class_types,
            native_types=native_types,
            table_mapping=table_mapping,
        )

    @classmethod
    def find_shimcache_win_xp(
        cls,
        context: interfaces.context.ContextInterface,
        kernel_module_name: str,
        shimcache_symbol_table: str,
    ) -> Iterator[shimcache.SHIM_CACHE_ENTRY]:
        """Attempts to find the shimcache in a Windows XP memory image

        :param context: The context to retrieve required elements (layers, symbol tables) from
        :param layer_name: The name of the memory layer on which to operate.
        :param kernel_symbol_table: The name of an existing symbol table containing the kernel symbols
        :param shimcache_symbol_table: The name of a symbol table containing the hand-crafted shimcache symbols
        """

        SHIM_NUM_ENTRIES_OFFSET = 0x8
        SHIM_MAX_ENTRIES = 0x60  # 96 max entries in XP shim cache
        SHIM_LRU_OFFSET = 0x10
        SHIM_HEADER_SIZE = 0x190
        SHIM_CACHE_ENTRY_SIZE = 0x228

        seen = set()

        for process in pslist.PsList.list_processes(context, kernel_module_name):
            pid = process.UniqueProcessId
            vollog.debug("checking process %d", pid)
            for vad in vadinfo.VadInfo.list_vads(
                process, lambda x: x.get_tag() == b"Vad " and x.Protection == 4
            ):
                try:
                    proc_layer_name = process.add_process_layer()
                    proc_layer = context.layers[proc_layer_name]
                except exceptions.InvalidAddressException:
                    continue

                try:
                    if proc_layer.read(vad.get_start(), 4) != b"\xef\xbe\xad\xde":
                        if pid == 624:
                            vollog.debug("VAD magic bytes don't match DEADBEEF")
                        continue
                except exceptions.InvalidAddressException:
                    continue

                num_entries = context.object(
                    shimcache_symbol_table + constants.BANG + "unsigned int",
                    proc_layer_name,
                    vad.get_start() + SHIM_NUM_ENTRIES_OFFSET,
                )

                vollog.debug(f"Found {num_entries} shimcache entries")

                if num_entries > SHIM_MAX_ENTRIES:
                    continue

                cache_idx_ptr = vad.get_start() + SHIM_LRU_OFFSET

                for _ in range(num_entries):
                    cache_idx_val = proc_layer.context.object(
                        shimcache_symbol_table + constants.BANG + "unsigned long",
                        proc_layer_name,
                        cache_idx_ptr,
                    )

                    cache_idx_ptr += 4

                    if cache_idx_val > SHIM_MAX_ENTRIES - 1:
                        continue

                    shim_entry_offset = (
                        vad.get_start()
                        + SHIM_HEADER_SIZE
                        + (SHIM_CACHE_ENTRY_SIZE * cache_idx_val)
                    )

                    if not proc_layer.is_valid(shim_entry_offset):
                        continue

                    physical_addr = proc_layer.translate(shim_entry_offset)

                    if physical_addr in seen:
                        continue

                    shim_entry = proc_layer.context.object(
                        shimcache_symbol_table + constants.BANG + "SHIM_CACHE_ENTRY",
                        proc_layer_name,
                        shim_entry_offset,
                    )
                    if not proc_layer.is_valid(shim_entry.vol.offset):
                        continue
                    if not shim_entry.is_valid():
                        continue

                    seen.add(physical_addr)

                    yield shim_entry

    @classmethod
    def find_shimcache_win_2k3_to_7(
        cls,
        context: interfaces.context.ContextInterface,
        config_path: str,
        kernel_module_name: str,
        shimcache_symbol_table: str,
    ) -> Iterator[shimcache.SHIM_CACHE_ENTRY]:
        """Implements the algorithm to search for the shim cache on Windows 2000
        (x64) through Windows 7 / 2008 R2. The algorithm consists of the following:

        1) Find the NT kernel module's .data and PAGE sections
        2) Iterate over every 4/8 bytes (depending on OS bitness) in the .data
           section and test for the following:
           a) offset represents a valid RTL_AVL_TABLE object
           b) RTL_AVL_TABLE is preceded by an ERESOURCE object
           c) RTL_AVL_TABLE is followed by the beginning of the SHIM LRU list

        :param context: The context to retrieve required elements (layers, symbol tables) from
        :param layer_name: The name of the memory layer on which to operate.
        :param kernel_symbol_table: The name of an existing symbol table containing the kernel symbols
        :param shimcache_symbol_table: The name of a symbol table containing the hand-crafted shimcache symbols
        """

        kernel = context.modules[kernel_module_name]

        data_sec = cls.get_module_section_range(
            context,
            config_path,
            kernel_module_name,
            cls.NT_KRNL_MODS,
            ".data",
        )
        mod_page = cls.get_module_section_range(
            context,
            config_path,
            kernel_module_name,
            cls.NT_KRNL_MODS,
            "PAGE",
        )

        # We require both in order to accurately handle AVL table
        if not (data_sec and mod_page):
            return

        data_sec_offset, data_sec_size = data_sec
        mod_page_offset, mod_page_size = mod_page

        addr_size = (
            8
            if symbols.symbol_table_is_64bit(
                context=context, symbol_table_name=kernel.symbol_table_name
            )
            else 4
        )

        shim_head = None
        for offset in range(
            data_sec_offset, data_sec_offset + data_sec_size, addr_size
        ):
            shim_head = cls.try_get_shim_head_at_offset(
                context,
                shimcache_symbol_table,
                kernel_module_name,
                mod_page_offset,
                mod_page_offset + mod_page_size,
                offset,
            )

            if shim_head:
                break

        if not shim_head:
            return

        yield from shim_head.ListEntry.to_list(
            shimcache_symbol_table + constants.BANG + "SHIM_CACHE_ENTRY", "ListEntry"
        )

    @classmethod
    def try_get_shim_head_at_offset(
        cls,
        context: interfaces.context.ContextInterface,
        shimcache_symbol_table: str,
        kernel_module_name: str,
        mod_page_start: int,
        mod_page_end: int,
        offset: int,
    ) -> Optional[shimcache.SHIM_CACHE_ENTRY]:
        """Attempts to construct a SHIM_CACHE_HEAD within a layer of the given context,
        using the provided offset within that layer, as well as the start and end offsets
        of the kernel module's `PAGE` section start and end offsets.

        If a number of validity checks are passed, this method will return the `SHIM_CACHE_HEAD`
        object. Otherwise, `None` is returned.
        """

        kernel = context.modules[kernel_module_name]

        # Check RTL_AVL_TABLE at offset
        rtl_avl_table = context.object(
            shimcache_symbol_table + constants.BANG + "_RTL_AVL_TABLE",
            kernel.layer_name,
            offset,
        )
        if not rtl_avl_table.is_valid(mod_page_start, mod_page_end):
            return None

        vollog.debug(f"Candidate RTL_AVL_TABLE found at offset {offset:#x}")

        ersrc_size = context.symbol_space.get_type(
            kernel.symbol_table_name + constants.BANG + "_ERESOURCE"
        ).size
        ersrc_alignment = (
            0x20
            if symbols.symbol_table_is_64bit(
                context=context, symbol_table_name=kernel.symbol_table_name
            )
            else 0x10
            # 0x20 if context.symbol_space.get_type("pointer").size == 8 else 0x10
        )
        vollog.debug(
            f"ERESOURCE size: {ersrc_size:#x}, ERESOURCE alignment: {ersrc_alignment:#x}"
        )

        eresource_rel_off = ersrc_size + ((offset - ersrc_size) % ersrc_alignment)
        eresource_offset = offset - eresource_rel_off

        vollog.debug(f"Constructing ERESOURCE at {eresource_offset:#x}")
        eresource = context.object(
            kernel.symbol_table_name + constants.BANG + "_ERESOURCE",
            kernel.layer_name,
            eresource_offset,
        )
        if not eresource.is_valid():
            vollog.debug("ERESOURCE Invalid")
            return None

        shim_head_offset = offset + rtl_avl_table.vol.size

        if not context.layers[kernel.layer_name].is_valid(shim_head_offset):
            return None

        shim_head = context.object(
            shimcache_symbol_table + constants.BANG + "SHIM_CACHE_ENTRY",
            kernel.layer_name,
            shim_head_offset,
        )

        if not shim_head.is_valid():
            vollog.debug("shim head invalid")
            return None
        else:
            vollog.debug("returning shim head")
            return shim_head

    @classmethod
    def find_shimcache_win_8_or_later(
        cls,
        context: interfaces.context.ContextInterface,
        config_path: str,
        kernel_module_name: str,
        shimcache_symbol_table: str,
    ) -> Iterator[shimcache.SHIM_CACHE_ENTRY]:
        """Attempts to locate and yield shimcache entries from a Windows 8 or later memory image.

        :param context: The context to retrieve required elements (layers, symbol tables) from
        :param layer_name: The name of the memory layer on which to operate.
        :param kernel_symbol_table: The name of an existing symbol table containing the kernel symbols
        :param shimcache_symbol_table: The name of a symbol table containing the hand-crafted shimcache symbols
        """
        kernel = context.modules[kernel_module_name]

        is_8_1_or_later = versions.is_windows_8_1_or_later(
            context, kernel.symbol_table_name
        ) or versions.is_win10(context, kernel.symbol_table_name)

        module_names = ["ahcache.sys"] if is_8_1_or_later else cls.NT_KRNL_MODS
        vollog.debug(f"Searching for modules {module_names}")

        data_sec = cls.get_module_section_range(
            context,
            config_path,
            kernel_module_name,
            module_names,
            ".data",
        )
        mod_page = cls.get_module_section_range(
            context,
            config_path,
            kernel_module_name,
            module_names,
            "PAGE",
        )

        if not (data_sec and mod_page):
            return None

        mod_page_offset, mod_page_size = mod_page
        data_sec_offset, data_sec_size = data_sec

        # iterate over ahcache kernel module's .data section in search of *two* SHIM handles
        shim_heads = []

        vollog.debug(f"PAGE offset: {mod_page_offset:#x}")
        vollog.debug(f".data offset: {data_sec_offset:#x}")

        handle_type = context.symbol_space.get_type(
            shimcache_symbol_table + constants.BANG + "SHIM_CACHE_HANDLE"
        )
        for offset in range(
            data_sec_offset,
            data_sec_offset + data_sec_size,
            (
                8
                if symbols.symbol_table_is_64bit(
                    context=context, symbol_table_name=kernel.symbol_table_name
                )
                else 4
            ),
        ):
            vollog.debug(f"Building shim handle pointer at {offset:#x}")
            shim_handle = context.object(
                object_type=shimcache_symbol_table + constants.BANG + "pointer",
                layer_name=kernel.layer_name,
                subtype=handle_type,
                offset=offset,
            )

            if shim_handle.is_valid(mod_page_offset, mod_page_offset + mod_page_size):
                if shim_handle.head is not None:
                    vollog.debug(
                        f"Found valid shim handle @ {shim_handle.vol.offset:#x}"
                    )
                    shim_heads.append(shim_handle.head)
                if len(shim_heads) == 2:
                    break

        if len(shim_heads) != 2:
            vollog.debug("Failed to identify two valid SHIM_CACHE_HANDLE structures")
            return

        # On Windows 8 x64, the first cache contains the shim cache.
        # On Windows 8 x86, 8.1 x86/x64, and 10, the second cache contains the shim cache.
        if (
            not symbols.symbol_table_is_64bit(
                context=context, symbol_table_name=kernel.symbol_table_name
            )
            and not is_8_1_or_later
        ):
            valid_head = shim_heads[1]
        elif not is_8_1_or_later:
            valid_head = shim_heads[0]
        else:
            valid_head = shim_heads[1]

        for shim_entry in valid_head.ListEntry.to_list(
            shimcache_symbol_table + constants.BANG + "SHIM_CACHE_ENTRY", "ListEntry"
        ):
            if shim_entry.is_valid():
                yield shim_entry

    def _generator(self):
        kernel = self.context.modules[self.config["kernel"]]

        shimcache_table_name = self.create_shimcache_table(
            self.context, kernel.symbol_table_name, self.config_path
        )

        c = count()

        if versions.is_windows_8_or_later(self._context, kernel.symbol_table_name):
            vollog.info("Finding shimcache entries for Windows 8.0+")
            entries = self.find_shimcache_win_8_or_later(
                self.context,
                self.config_path,
                self.config["kernel"],
                shimcache_table_name,
            )

        elif (
            versions.is_2003(self.context, kernel.symbol_table_name)
            or versions.is_vista_or_later(self.context, kernel.symbol_table_name)
            or versions.is_windows_7(self.context, kernel.symbol_table_name)
        ):
            vollog.info("Finding shimcache entries for Windows 2k3/Vista/7")
            entries = self.find_shimcache_win_2k3_to_7(
                self.context,
                self.config_path,
                self.config["kernel"],
                shimcache_table_name,
            )

        elif versions.is_windows_xp_sp2(
            self._context, kernel.symbol_table_name
        ) or versions.is_windows_xp_sp3(self.context, kernel.symbol_table_name):
            vollog.info("Finding shimcache entries for WinXP")
            entries = self.find_shimcache_win_xp(
                self._context,
                self.config["kernel"],
                shimcache_table_name,
            )
        else:
            vollog.warn("Cannot parse shimcache entries for this version of Windows")
            return

        for entry in entries:
            try:
                vollog.debug(f"SHIM_CACHE_ENTRY type: {entry.__class__}")
                shim_entry = (
                    entry.last_modified,
                    entry.last_update,
                    entry.exec_flag,
                    (
                        format_hints.Hex(entry.file_size)
                        if isinstance(entry.file_size, int)
                        else entry.file_size
                    ),
                    entry.file_path,
                )
            except exceptions.InvalidAddressException:
                continue

            yield (
                0,
                (next(c), *shim_entry),
            )

    def run(self):
        return renderers.TreeGrid(
            [
                ("Order", int),
                ("Last Modified", datetime),
                ("Last Update", datetime),
                ("Exec Flag", bool),
                ("File Size", format_hints.Hex),
                ("File Path", str),
            ],
            self._generator(),
        )

    @classmethod
    def get_module_section_range(
        cls,
        context: interfaces.context.ContextInterface,
        config_path: str,
        kernel_module_name: str,
        module_list: List[str],
        section_name: str,
    ) -> Optional[Tuple[int, int]]:
        """Locates the size and offset of the first found module section
        specified by name from the list of modules.

        :param context: The context to operate on
        :param layer_name: The memory layer to read from
        :param module_list: A list of module names to search for the given section
        :param section_name: The name of the section to search for.

        :return: The offset and size of the module, if found; Otherwise, returns `None`
        """

        krnl_mod = None
        for module in modules.Modules.list_modules(context, kernel_module_name):
            try:
                if module.BaseDllName.String in module_list:
                    krnl_mod = module
                    break
            except exceptions.InvalidAddressException as exc:
                vollog.warning(
                    f"Failed to get kernel module due to {exc.__class__.__name__}: {exc.invalid_address:#x}"
                )

        if krnl_mod is None:
            vollog.warning("Failed to find kernel module")
            return None

        kernel = context.modules[kernel_module_name]

        pe_table_name = intermed.IntermediateSymbolTable.create(
            context,
            interfaces.configuration.path_join(config_path, "pe"),
            "windows",
            "pe",
            class_types=pe.class_types,
        )

        # code taken from Win32KBase._section_chunks (win32_core.py)
        dos_header = context.object(
            pe_table_name + constants.BANG + "_IMAGE_DOS_HEADER",
            kernel.layer_name,
            offset=krnl_mod.DllBase,
        )

        if not dos_header:
            return None

        nt_header = dos_header.get_nt_header()

        try:
            section = next(
                sec
                for sec in nt_header.get_sections()
                if section_name.lower() == array_to_string(sec.Name).lower()
            )
        except StopIteration:
            return None

        section_offset = krnl_mod.DllBase + section.VirtualAddress
        section_size = section.Misc.VirtualSize

        return section_offset, section_size



# This file is Copyright 2019 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#

from typing import Iterable

from volatility3.framework import renderers, interfaces, exceptions
from volatility3.framework.configuration import requirements
from volatility3.framework.renderers import format_hints
from volatility3.plugins.windows import poolscanner


class FileScan(interfaces.plugins.PluginInterface):
    """Scans for file objects present in a particular windows memory image."""

    _required_framework_version = (2, 0, 0)
    _version = (2, 0, 0)

    @classmethod
    def get_requirements(cls):
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Windows kernel",
                architectures=["Intel32", "Intel64"],
            ),
            requirements.VersionRequirement(
                name="poolscanner", component=poolscanner.PoolScanner, version=(3, 0, 0)
            ),
        ]

    @classmethod
    def scan_files(
        cls,
        context: interfaces.context.ContextInterface,
        kernel_module_name: str,
    ) -> Iterable[interfaces.objects.ObjectInterface]:
        """Scans for file objects using the poolscanner module and constraints.

        Args:
            context: The context to retrieve required elements (layers, symbol tables) from
            kernel_module_name: The name of the module for the kernel

        Returns:
            A list of File objects as found from the `layer_name` layer based on File pool signatures
        """

        kernel = context.modules[kernel_module_name]

        constraints = poolscanner.PoolScanner.builtin_constraints(
            kernel.symbol_table_name, [b"Fil\xe5", b"File"]
        )

        for result in poolscanner.PoolScanner.generate_pool_scan(
            context, kernel_module_name, constraints
        ):
            _constraint, mem_object, _header = result
            yield mem_object

    def _generator(self):
        for fileobj in self.scan_files(self.context, self.config["kernel"]):
            try:
                file_name = fileobj.FileName.String
            except exceptions.InvalidAddressException:
                continue

            yield (0, (format_hints.Hex(fileobj.vol.offset), file_name))

    def run(self):
        return renderers.TreeGrid(
            [("Offset", format_hints.Hex), ("Name", str)],
            self._generator(),
        )



# This file is Copyright 2021 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#

import logging

from volatility3.framework import interfaces, renderers
from volatility3.framework.configuration import requirements
from volatility3.framework.renderers import format_hints
from volatility3.plugins.linux import pslist

vollog = logging.getLogger(__name__)


class PsTree(interfaces.plugins.PluginInterface):
    """Plugin for listing processes in a tree based on their parent process ID."""

    _required_framework_version = (2, 13, 0)
    _version = (1, 1, 1)

    @classmethod
    def get_requirements(cls):
        # Since we're calling the plugin, make sure we have the plugin's requirements
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Linux kernel",
                architectures=["Intel32", "Intel64"],
            ),
            requirements.VersionRequirement(
                name="pslist", component=pslist.PsList, version=(4, 0, 0)
            ),
            requirements.ListRequirement(
                name="pid",
                description="Filter on specific process IDs",
                element_type=int,
                optional=True,
            ),
            requirements.BooleanRequirement(
                name="threads",
                description="Include user threads",
                optional=True,
                default=False,
            ),
            requirements.BooleanRequirement(
                name="decorate_comm",
                description="Show `user threads` comm in curly brackets, and `kernel threads` comm in square brackets",
                optional=True,
                default=False,
            ),
        ]

    def find_level(self, pid: int) -> None:
        """Finds how deep the PID is in the tasks hierarchy.

        Args:
            pid: PID to find the level in the hierarchy
        """
        seen_ppids = set()
        seen_offsets = set()

        level = 0
        proc = self._tasks.get(pid)

        while proc:
            # we don't want swapper in the tree
            if proc.pid == 0:
                break

            if proc.is_thread_group_leader:
                parent_pid = proc.get_parent_pid()
            else:
                parent_pid = proc.tgid

            if parent_pid in seen_ppids or proc.vol.offset in seen_offsets:
                break

            # only pid 1 (init/systemd) or 2 (kthreadd) should have swapper as a parent
            # any other process with a ppid of 0 is smeared or terminated
            if parent_pid == 0 and proc.pid > 2:
                vollog.debug(
                    "Smeared process with parent PID of 0 and PID greater than 2 ({proc.pid}) is being skipped."
                )
                break

            seen_ppids.add(parent_pid)
            seen_offsets.add(proc.vol.offset)

            child_list = self._children.setdefault(parent_pid, set())
            child_list.add(proc.pid)

            proc = self._tasks.get(parent_pid)

            level += 1

        self._levels[pid] = level

    def _generator(
        self,
        tasks: list,
        decorate_comm: bool = False,
    ):
        """Generates the tasks hierarchy tree.

        Args:
            tasks: A list of task objects to be displayed
            decorate_comm: If True, it decorates the comm string of
                            - User threads: in curly brackets,
                            - Kernel threads: in square brackets
                           Defaults to False.
        Yields:
            Each rows
        """

        self._tasks = {}
        self._levels = {}
        self._children = {}

        for proc in tasks:
            self._tasks[proc.pid] = proc

        # Build the child/level maps
        for pid in self._tasks:
            self.find_level(pid)

        def yield_processes(pid):
            task = self._tasks[pid]

            task_fields = pslist.PsList.get_task_fields(task, decorate_comm)
            fields = (
                format_hints.Hex(task_fields.offset),
                task_fields.user_pid,
                task_fields.user_tid,
                task_fields.user_ppid,
                task_fields.name,
            )
            yield (self._levels[task_fields.user_tid] - 1, fields)

            seen_children = set()

            for child_pid in sorted(self._children.get(task_fields.user_tid, [])):
                if child_pid in seen_children:
                    break
                seen_children.add(child_pid)

                yield from yield_processes(child_pid)

        seen_processes = set()

        for pid, level in self._levels.items():
            if level == 1:
                for fields in yield_processes(pid):
                    pid = fields[1]
                    if pid in seen_processes:
                        break
                    seen_processes.add(pid)

                    yield fields

    def run(self):
        filter_func = pslist.PsList.create_pid_filter(self.config.get("pid", None))
        include_threads = self.config.get("threads")
        decorate_comm = self.config.get("decorate_comm")

        return renderers.TreeGrid(
            [
                ("OFFSET (V)", format_hints.Hex),
                ("PID", int),
                ("TID", int),
                ("PPID", int),
                ("COMM", str),
            ],
            self._generator(
                pslist.PsList.list_tasks(
                    self.context,
                    self.config["kernel"],
                    filter_func=filter_func,
                    include_threads=include_threads,
                ),
                decorate_comm=decorate_comm,
            ),
        )



# This file is Copyright 2025 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#
import logging
from volatility3.framework import interfaces, deprecation
from collections import namedtuple
from volatility3.plugins.windows.malware import direct_system_calls

vollog = logging.getLogger(__name__)

# Full details on the techniques used in these plugins to detect EDR-evading malware
# can be found in our 20 page whitepaper submitted to DEFCON along with the presentation
# https://www.volexity.com/wp-content/uploads/2024/08/Defcon24_EDR_Evasion_Detection_White-Paper_Andrew-Case.pdf

syscall_finder_type = namedtuple(
    "syscall_finder_type",
    [
        "get_syscall_target_address",
        "wants_syscall_inst",
        "rule_str",
        "invalid_ops",
        "termination_ops",
    ],
)


class DirectSystemCalls(
    interfaces.plugins.PluginInterface,
    deprecation.PluginRenameClass,
    replacement_class=direct_system_calls.DirectSystemCalls,
    removal_date="2026-06-07",
):
    """Detects the Direct System Call technique used to bypass EDRs (deprecated)."""

    _required_framework_version = (2, 4, 0)

    # 2.0.0 - changes signature of `get_tasks_to_scan`
    _version = (2, 0, 0)

    # DLLs that are expected to host system call invocations
    valid_syscall_handlers = ("ntdll.dll", "win32u.dll")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.syscall_finder = syscall_finder_type(
            # for direct system calls, we find the `syscall` instruction directly, so we already know the address
            None,
            # yes, we want the syscall instruction present as it is what this technique looks for
            True,
            # regex to find "\x0f\x05" (syscall) followed later by "\xc3" (ret)
            # we allow spacing in between to break naive anti-analysis forms (e.g., TarTarus Gate)
            # Standard techniques, such as HellsGate, look like:
            # mov r10, rcx
            # mov eax, <system call number>
            # syscall
            # ret
            "/\\x0f\\x05[^\\xc3]{,24}\\xc3/",
            # any of these will not be in a workable, malicious direct system call block
            ["jmp", "call", "leave", "int3"],
            # the expected form is to end with a "ret" back to the calling code
            ["ret"],
        )



# This file is Copyright 2019 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#

from volatility3.framework import renderers, interfaces, exceptions
from volatility3.framework.configuration import requirements
from volatility3.framework.objects import utility
from volatility3.framework.renderers import format_hints
from volatility3.plugins.mac import pslist
from typing import Callable, Generator, Type, Optional
import logging

vollog = logging.getLogger(__name__)


class Maps(interfaces.plugins.PluginInterface):
    """Lists process memory ranges that potentially contain injected code."""

    _required_framework_version = (2, 0, 0)
    _version = (1, 1, 0)
    MAXSIZE_DEFAULT = 1024 * 1024 * 1024  # 1 Gb

    @classmethod
    def get_requirements(cls):
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Kernel module for the OS",
                architectures=["Intel32", "Intel64"],
            ),
            requirements.VersionRequirement(
                name="pslist", component=pslist.PsList, version=(3, 0, 0)
            ),
            requirements.ListRequirement(
                name="pid",
                description="Filter on specific process IDs",
                element_type=int,
                optional=True,
            ),
            requirements.BooleanRequirement(
                name="dump",
                description="Extract listed memory segments",
                default=False,
                optional=True,
            ),
            requirements.ListRequirement(
                name="address",
                description="Process virtual memory addresses to include "
                "(all other VMA sections are excluded). This can be any "
                "virtual address within the VMA section. Virtual addresses "
                "must be separated by a space.",
                element_type=int,
                optional=True,
            ),
            requirements.IntRequirement(
                name="maxsize",
                description="Maximum size for dumped VMA sections "
                "(all the bigger sections will be ignored)",
                default=cls.MAXSIZE_DEFAULT,
                optional=True,
            ),
        ]

    @classmethod
    def list_vmas(
        cls,
        task: interfaces.objects.ObjectInterface,
        filter_func: Callable[[interfaces.objects.ObjectInterface], bool] = lambda _: (
            True
        ),
    ) -> Generator[interfaces.objects.ObjectInterface, None, None]:
        """Lists the Virtual Memory Areas of a specific process.

        Args:
            task: task object from which to list the vma
            filter_func: Function to take a vma and return False if it should be filtered out

        Returns:
            Yields vmas based on the task and filtered based on the filter function
        """
        for vma in task.get_map_iter():
            if filter_func(vma):
                yield vma
            else:
                vollog.debug(
                    f"Excluded vma at offset {vma.vol.offset:#x} for pid {task.p_pid} due to filter_func"
                )

    @classmethod
    def vma_dump(
        cls,
        context: interfaces.context.ContextInterface,
        task: interfaces.objects.ObjectInterface,
        vm_start: int,
        vm_end: int,
        open_method: Type[interfaces.plugins.FileHandlerInterface],
        maxsize: int = MAXSIZE_DEFAULT,
    ) -> Optional[interfaces.plugins.FileHandlerInterface]:
        """Extracts the complete data for VMA as a FileInterface.

        Args:
            context: The context to retrieve required elements (layers, symbol tables) from
            task: an task_struct instance
            vm_start: The start virtual address from the vma to dump
            vm_end: The end virtual address from the vma to dump
            open_method: class to provide context manager for opening the file
            maxsize: Max size of VMA section (default MAXSIZE_DEFAULT)

        Returns:
            An open FileInterface object containing the complete data for the task or None in the case of failure
        """
        pid = task.p_pid

        try:
            proc_layer_name = task.add_process_layer()
        except exceptions.InvalidAddressException as excp:
            vollog.debug(
                f"Process {pid}: invalid address {excp.invalid_address} in layer {excp.layer_name}"
            )
            return None
        vm_size = vm_end - vm_start

        # check if vm_size is negative, this should never happen.
        if vm_size < 0:
            vollog.warning(
                f"Skip virtual memory dump for pid {pid} between {vm_start:#x}-{vm_end:#x} as {vm_size} is negative."
            )
            return None
        # check if vm_size is larger than the maxsize limit, and therefore is not saved out.
        if maxsize <= vm_size:
            vollog.warning(
                f"Skip virtual memory dump for pid {pid} between {vm_start:#x}-{vm_end:#x} as {vm_size} is larger than maxsize limit of {maxsize}"
            )
            return None
        proc_layer = context.layers[proc_layer_name]
        file_name = f"pid.{pid}.vma.{vm_start:#x}-{vm_end:#x}.dmp"
        try:
            file_handle = open_method(file_name)
            chunk_size = 1024 * 1024 * 10
            offset = vm_start
            while offset < vm_start + vm_size:
                to_read = min(chunk_size, vm_start + vm_size - offset)
                data = proc_layer.read(offset, to_read, pad=True)
                file_handle.write(data)
                offset += to_read
        except Exception as excp:
            vollog.debug(f"Unable to dump virtual memory {file_name}: {excp}")
            return None
        return file_handle

    def _generator(self, tasks):
        address_list = self.config.get("address", None)
        if not address_list:
            # do not filter as no address_list was supplied
            def vma_filter_func(_):
                return True

        else:
            # filter for any vm_start that matches the supplied address config
            def vma_filter_function(task: interfaces.objects.ObjectInterface) -> bool:
                addrs_in_vma = [
                    addr
                    for addr in address_list
                    if task.links.start <= addr <= task.links.end
                ]

                # if any of the user supplied addresses would fall within this vma return true
                return bool(addrs_in_vma)

            vma_filter_func = vma_filter_function

        for task in tasks:
            process_name = utility.array_to_string(task.p_comm)
            process_pid = task.p_pid

            for vma in self.list_vmas(task, filter_func=vma_filter_func):
                try:
                    vm_start = vma.links.start
                    vm_end = vma.links.end
                except AttributeError:
                    vollog.debug(
                        f"Unable to find the vm_start and vm_end for vma at {vma.vol.offset:#x} for pid {process_pid}"
                    )
                    continue

                path = vma.get_path(
                    self.context,
                    self.context.modules[self.config["kernel"]].symbol_table_name,
                )
                if path == "":
                    path = vma.get_special_path()

                file_output = "Disabled"
                if self.config["dump"]:
                    file_output = "Error outputting file"
                    file_handle = self.vma_dump(
                        self.context,
                        task,
                        vm_start,
                        vm_end,
                        self.open,
                        self.config["maxsize"],
                    )

                    if file_handle:
                        file_handle.close()
                        file_output = file_handle.preferred_filename

                yield (
                    0,
                    (
                        process_pid,
                        process_name,
                        format_hints.Hex(vm_start),
                        format_hints.Hex(vm_end),
                        vma.get_perms(),
                        path,
                        file_output,
                    ),
                )

    def run(self):
        filter_func = pslist.PsList.create_pid_filter(self.config.get("pid", None))
        list_tasks = pslist.PsList.get_list_tasks(
            self.config.get("pslist_method", pslist.PsList.pslist_methods[0])
        )

        return renderers.TreeGrid(
            [
                ("PID", int),
                ("Process", str),
                ("Start", format_hints.Hex),
                ("End", format_hints.Hex),
                ("Protection", str),
                ("Map Name", str),
                ("File output", str),
            ],
            self._generator(
                list_tasks(self.context, self.config["kernel"], filter_func=filter_func)
            ),
        )
